__author__ = 'Robbert Harms'
__date__ = '2018-08-01'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


return_code_labels = {
    0: ['default', 'no return code specified'],
    1: ['found zero', 'sum of squares/evaluation below underflow limit'],
    2: ['converged', 'the relative error in the sum of squares/evaluation is at most tol'],
    3: ['converged', 'the relative error of the parameter vector is at most tol'],
    4: ['converged', 'both errors are at most tol'],
    5: ['trapped', 'by degeneracy; increasing epsilon might help'],
    6: ['exhausted', 'number of function calls exceeding preset patience'],
    7: ['failed', 'ftol<tol: cannot reduce sum of squares any further'],
    8: ['failed', 'xtol<tol: cannot improve approximate solution any further'],
    9: ['failed', 'gtol<tol: cannot improve approximate solution any further'],
    10: ['NaN', 'Function value is not-a-number or infinite'],
    11: ['exhausted', 'temperature decreased to 0.0']
}


class OptimizeResults(dict):
    """Represents the optimization results.

    Attributes:
        x (ndarray): the optimized parameter maps, an (d, p) array with for d problems a value for every p parameters
        status (ndarray): the return codes, an (d,) vector with for d problems the status return code
    """
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())



import unittest

import numpy as np
import pyopencl as cl

from mot.lib.utils import device_type_from_string, device_supports_double, get_float_type_def, is_scalar, \
    all_elements_equal, get_single_value, topological_sort

__author__ = 'Robbert Harms'
__date__ = "2017-03-28"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class test_device_type_from_string(unittest.TestCase):

    def test_gpu(self):
        assert(device_type_from_string('GPU') == cl.device_type.GPU)

    def test_cpu(self):
        assert(device_type_from_string('CPU') == cl.device_type.CPU)

    def test_accelerator(self):
        assert(device_type_from_string('ACCELERATOR') == cl.device_type.ACCELERATOR)

    def test_custom(self):
        assert(device_type_from_string('CUSTOM') == cl.device_type.CUSTOM)

    def test_none(self):
        assert(device_type_from_string('') is None)


class test_device_supports_double(unittest.TestCase):

    def test_has_double(self):
        for platform in cl.get_platforms():
            for device in platform.get_devices():
                has_double = device.get_info(cl.device_info.DOUBLE_FP_CONFIG) == 63
                assert(device_supports_double(device) == has_double)


class test_is_scalar(unittest.TestCase):

    def test_is_not_scalar(self):
        self.assertFalse(is_scalar(np.zeros((2, 2))))

    def test_is_scalar(self):
        self.assertTrue(is_scalar(np.zeros((1, ))[:, None]))
        self.assertTrue(is_scalar(np.zeros((1,))))
        self.assertTrue(is_scalar(-1))
        self.assertTrue(is_scalar(0))
        self.assertTrue(is_scalar(1))
        self.assertTrue(is_scalar(-1.0))
        self.assertTrue(is_scalar(0.0))
        self.assertTrue(is_scalar(1.0))


class test_all_elements_equal(unittest.TestCase):

    def test_scalar(self):
        self.assertTrue(all_elements_equal(np.zeros((1,))[:, None]))
        self.assertTrue(all_elements_equal(np.zeros((1,))))
        self.assertTrue(all_elements_equal(-1))
        self.assertTrue(all_elements_equal(0))
        self.assertTrue(all_elements_equal(1))
        self.assertTrue(all_elements_equal(-1.0))
        self.assertTrue(all_elements_equal(0.0))
        self.assertTrue(all_elements_equal(1.0))

    def test_false(self):
        self.assertFalse(all_elements_equal(np.random.rand(2, 2)))

    def test_matrix(self):
        self.assertTrue(all_elements_equal(np.zeros((2,))))
        self.assertTrue(all_elements_equal(np.zeros((2, 3))))
        self.assertTrue(all_elements_equal(np.zeros((2, 3, 4))))


class test_get_single_value(unittest.TestCase):

    def test_exception(self):
        self.assertRaises(ValueError, get_single_value, np.random.rand(2, 2))

    def test_true(self):
        self.assertTrue(get_single_value(np.ones((2, 2))[:, None]) == 1)
        self.assertTrue(get_single_value(np.zeros((1,))[:, None]) == 0)
        self.assertTrue(get_single_value(np.zeros((1,))) == 0)
        self.assertTrue(get_single_value(-1) == -1)
        self.assertTrue(get_single_value(0) == 0)
        self.assertTrue(get_single_value(1) == 1)
        self.assertTrue(get_single_value(-1.0) == -1.0)
        self.assertTrue(get_single_value(0.0) == 0.0)
        self.assertTrue(get_single_value(1.0) == 1.0)


class test_topological_sort(unittest.TestCase):

    def test_auto_dependency(self):
        circular = {'a': ('a',), 'b': ('a',)}
        self.assertRaises(ValueError, topological_sort, circular)

    def test_cyclic_dependency(self):
        circular = {'a': ('b',), 'b': ('a',)}
        self.assertRaises(ValueError, topological_sort, circular)

    def test_sorting(self):
        data = {'a': (), 'm': ('c',), 'e': ('m',), '!': ('a', 'e', 'c')}
        assert(list(topological_sort(data)) == ['a', 'c', 'm', 'e', '!'])

    def test_unsortables(self):
        class A():
            pass

        class B():
            pass

        a = A()
        b = B()
        data = {'a': (a, b), 'b': ('a',)}

        assert(len(list(topological_sort(data))) == 4)

    def test_empty_input(self):
        data = {}
        self.assertFalse(topological_sort(data))



from mot.library_functions.base import SimpleCLLibrary

__author__ = 'Robbert Harms'
__date__ = '2018-05-07'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class polevl(SimpleCLLibrary):

    def __init__(self):
        """Routines for computing polynomials.

        Copied from Scipy (https://github.com/scipy/scipy/blob/master/scipy/special/cephes/polevl.h), 05-05-2018.

        Evaluates polynomial of degree N:

                            2          N
        y  =  C  + C x + C x  +...+ C x
               0      1     2          N

        Coefficients are stored in reverse order:

        coef[0] = C  , ..., coef[N] = C  .
                   N                   0
        """
        super().__init__('''
            double polevl(double x, double* coef, int N){
                double ans;
                double *p;
                int i;

                p = coef;
                ans = *p++;
                i = N;

                do
                    ans = ans * x  +  *p++;
                while( --i );

                return ans;
            }
        ''')


class p1evl(SimpleCLLibrary):

    def __init__(self):
        """Routines for computing polynomials when coefficient of x^N is 1.0.

        Copied from Scipy (https://github.com/scipy/scipy/blob/master/scipy/special/cephes/polevl.h), 05-05-2018.

        Evaluates polynomial of degree N:

                            2          N
        y  =  C  + C x + C x  +...+ C x
               0      1     2          N

        Coefficients are stored in reverse order:

        coef[0] = C  , ..., coef[N] = C  .
                   N                   0

        In contrast to ``polevl``, this function assumes that coef[N] = 1.0 and is omitted from the array.
        Its calling arguments are otherwise the same as polevl().
        """
        super(p1evl, self).__init__('''
            double p1evl(double x, double* coef, int N){
                double ans;
                double *p;
                int i;

                p = coef;
                ans = x + *p++;
                i = N-1;

                do
                    ans = ans * x  + *p++;
                while( --i );

                return( ans );
            }
        ''')


class ratevl(SimpleCLLibrary):

    def __init__(self):
        """Evaluates a rational integral.

        Copied from Scipy (https://github.com/scipy/scipy/blob/master/scipy/special/cephes/polevl.h), 2018-05-07.
        """
        super(ratevl, self).__init__('''
            double ratevl(double x, double* num, int M, double* denom, int N){
                int i, dir;
                double y, num_ans, denom_ans;
                double absx = fabs(x);
                double *p;

                if (absx > 1) {
                    /* Evaluate as a polynomial in 1/x. */
                    dir = -1;
                    p = num + M;
                    y = 1 / x;
                } else {
                    dir = 1;
                    p = num;
                    y = x;
                }

                /* Evaluate the numerator */
                num_ans = *p;
                p += dir;
                for (i = 1; i <= M; i++) {
                    num_ans = num_ans * y + *p;
                    p += dir;
                }

                /* Evaluate the denominator */
                if (absx > 1) {
                    p = denom + N;
                } else {
                    p = denom;
                }

                denom_ans = *p;
                p += dir;
                for (i = 1; i <= N; i++) {
                    denom_ans = denom_ans * y + *p;
                    p += dir;
                }

                if (absx > 1) {
                    i = N - M;
                    return pow(x, i) * num_ans / denom_ans;
                } else {
                    return num_ans / denom_ans;
                }
            }
        ''')




import logging
from contextlib import contextmanager

from mot.lib.cl_function import SimpleCLFunction
from mot.lib.cl_runtime_info import CLRuntimeInfo
from mot.library_functions import Rand123
from mot.lib.utils import split_in_batches
from mot.lib.kernel_data import Scalar, LocalMemory, Array, \
    Zeros
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractSampler(object):

    def __init__(self, model, starting_positions, cl_runtime_info=None, **kwargs):
        """Abstract base class for sample routines.

        Sampling routines implementing this interface should be stateful objects that, for the given model, keep track
        of the sample state over multiple calls to :meth:`sample`.

        Args:
            model (SampleModelInterface): the model to sample.
            starting_positions (ndarray): the starting positions for the sampler. Should be a two dimensional matrix
                with for every modeling instance (first dimension) and every parameter (second dimension) a value.
        """
        self._cl_runtime_info = cl_runtime_info or CLRuntimeInfo()
        self._logger = logging.getLogger(__name__)
        self._model = model
        self._starting_positions = starting_positions
        if len(starting_positions.shape) < 2:
            self._starting_positions = self._starting_positions[..., None]
        self._nmr_problems = self._starting_positions.shape[0]
        self._nmr_params = self._starting_positions.shape[1]
        self._sampling_index = 0
        self._current_chain_position = np.require(np.copy(self._starting_positions),
                                                  requirements=['C', 'A', 'O', 'W'],
                                                  dtype=self._cl_runtime_info.mot_float_dtype)
        self._rng_state = np.random.uniform(low=np.iinfo(np.uint32).min, high=np.iinfo(np.uint32).max + 1,
                                            size=(self._nmr_problems, 6)).astype(np.uint32)

    def set_cl_runtime_info(self, cl_runtime_info):
        """Update the CL runtime information.

        Args:
            cl_runtime_info (mot.lib.cl_runtime_info.CLRuntimeInfo): the new runtime information
        """
        self._cl_runtime_info = cl_runtime_info

    def sample(self, nmr_samples, burnin=0, thinning=1):
        """Take additional samples from the given model using this sampler.

        This method can be called multiple times in which the sample state is stored in between.

        Args:
            nmr_samples (int): the number of samples to return
            burnin (int): the number of samples to discard before returning samples
            thinning (int): how many sample we wait before storing a new one. This will draw extra samples such that
                    the total number of samples generated is ``nmr_samples * (thinning)`` and the number of samples
                    stored is ``nmr_samples``. If set to one or lower we store every sample after the burn in.

        Returns:
            SamplingOutput: the sample output object
        """
        if not thinning or thinning < 1:
            thinning = 1
        if not burnin or burnin < 0:
            burnin = 0

        with self._logging(nmr_samples, burnin, thinning):
            if burnin > 0:
                for batch_start, batch_end in split_in_batches(burnin, max(1000 // thinning, 100)):
                    self._sample(batch_end - batch_start, return_output=False)
            if nmr_samples > 0:
                outputs = []
                for batch_start, batch_end in split_in_batches(nmr_samples, max(1000 // thinning, 100)):
                    outputs.append(self._sample(batch_end - batch_start, thinning=thinning))
                return SimpleSampleOutput(*[np.concatenate([o[ind] for o in outputs], axis=-1) for ind in range(3)])

    def _sample(self, nmr_samples, thinning=1, return_output=True):
        """Sample the given number of samples with the given thinning.

        If ``return_output`` we will return the samples, log likelihoods and log priors. If not, we will advance the
        state of the sampler without returning storing the samples.

        Args:
            nmr_samples (int): the number of iterations to advance the sampler
            thinning (int): the thinning to apply
            return_output (boolean): if we should return the output

        Returns:
            None or tuple: if ``return_output`` is True three ndarrays as (samples, log_likelihoods, log_priors)
        """
        kernel_data = self._get_kernel_data(nmr_samples, thinning, return_output)
        sample_func = self._get_compute_func(nmr_samples, thinning, return_output)
        sample_func.evaluate({'data': kernel_data}, nmr_instances=self._nmr_problems, use_local_reduction=True,
                             cl_runtime_info=self._cl_runtime_info)
        self._sampling_index += nmr_samples * thinning
        self._readout_kernel_data(kernel_data)
        if return_output:
            return (kernel_data['_samples'].get_data(),
                    kernel_data['_log_likelihoods'].get_data(),
                    kernel_data['_log_priors'].get_data())

    def _get_kernel_data(self, nmr_samples, thinning, return_output):
        """Get the kernel data we will input to the MCMC sampler.

        By default, this will take the kernel data from the model and add to that the items:

        * _nmr_iterations: the number of iterations to sample
        * _iteration_offset: the current sample index, that is, the offset to the given number of iterations
        * _rng_state: the random number generator state
        * _current_chain_position: the current position of the sampled chain
        * _log_likelihood_tmp: an OpenCL local array for combining the log likelihoods

        Additionally, if ``return_output`` is True, we add to that the arrays:

        * _samples: for the samples
        * _log_likelihoods: for storing the log likelihoods
        * _log_priors: for storing the priors

        Args:
            nmr_samples (int): the number of samples we will draw
            thinning (int): the thinning factor we want to use
            return_output (boolean): if the kernel should return output

        Returns:
            dict[str: mot.lib.utils.KernelData]: the kernel input data
        """
        kernel_data = dict(self._model.get_kernel_data())
        kernel_data.update({
            '_nmr_iterations': Scalar(nmr_samples * thinning, ctype='ulong'),
            '_iteration_offset': Scalar(self._sampling_index, ctype='ulong'),
            '_rng_state': Array(self._rng_state, 'uint', is_writable=True, ensure_zero_copy=True),
            '_current_chain_position': Array(self._current_chain_position, 'mot_float_type',
                                             is_writable=True, ensure_zero_copy=True),
            '_log_likelihood_tmp': LocalMemory('double')
        })

        if return_output:
            kernel_data.update({
                '_samples': Zeros((self._nmr_problems, self._nmr_params, nmr_samples),
                                                       'mot_float_type'),
                '_log_likelihoods': Zeros((self._nmr_problems, nmr_samples), 'mot_float_type'),
                '_log_priors': Zeros((self._nmr_problems, nmr_samples), 'mot_float_type'),
            })
        return kernel_data

    def _readout_kernel_data(self, kernel_data):
        """Readout the kernel data and update the sampler state with the state from the compute device.

        Args:
            kernel_data (dict[str: mot.lib.utils.KernelData]): the kernel data from which to read the output
        """
        pass

    def _get_compute_func(self, nmr_samples, thinning, return_output):
        """Get the MCMC algorithm as a computable function.

        Args:
            nmr_samples (int): the number of samples we will draw
            thinning (int): the thinning factor we want to use
            return_output (boolean): if the kernel should return output

        Returns:
            mot.lib.cl_function.CLFunction: the compute function
        """
        kernel_source = self._get_state_update_cl_func(nmr_samples, thinning, return_output)

        cl_func = '''
            void compute(mot_data_struct* data){
                bool is_first_work_item = get_local_id(0) == 0;
    
                rand123_data rand123_rng_data = rand123_initialize_data((uint[]){
                    data->_rng_state[0], data->_rng_state[1], data->_rng_state[2], data->_rng_state[3], 
                    data->_rng_state[4], data->_rng_state[5], 0, 0});
                void* rng_data = (void*)&rand123_rng_data;
    
                local mot_float_type current_position[''' + str(self._nmr_params) + '''];
                local double current_likelihood;
                local mot_float_type current_prior;
    
                if(is_first_work_item){
                    for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        current_position[i] = data->_current_chain_position[i];
                    }
                    current_prior = _computeLogPrior(data, current_position);
                }
                barrier(CLK_LOCAL_MEM_FENCE);
    
                current_likelihood = _computeLogLikelihood(data, current_position, data->_log_likelihood_tmp);
    
                for(ulong i = 0; i < data->_nmr_iterations; i++){
        '''
        if return_output:
            cl_func += '''
                    if(is_first_work_item){
                        if(i % ''' + str(thinning) + ''' == 0){
    
                            data->_log_likelihoods[i / ''' + str(thinning) + '''] = current_likelihood;
                            data->_log_priors[i / ''' + str(thinning) + '''] = current_prior;
    
                            for(uint j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                                data->_samples[(ulong)(i / ''' + str(thinning) + ''') // remove the interval
                                        + j * ''' + str(nmr_samples) + '''  // parameter index
                                ] = current_position[j];
                            }
                        }
                    }
        '''
        cl_func += '''
                    _advanceSampler(data, i + data->_iteration_offset, rng_data, 
                                    current_position, &current_likelihood, &current_prior, data->_log_likelihood_tmp);
                }

                if(is_first_work_item){
                    for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        data->_current_chain_position[i] = current_position[i];
                    }

                    uint state[8];
                    rand123_data_to_array(rand123_rng_data, state);
                    for(uint i = 0; i < 6; i++){
                        data->_rng_state[i] = state[i];
                    }
                }
            }
        '''
        return SimpleCLFunction.from_string(
            cl_func,
            dependencies=[Rand123(), self._get_log_prior_cl_func(), self._get_log_likelihood_cl_func()],
            cl_extra=kernel_source)

    def _get_state_update_cl_func(self, nmr_samples, thinning, return_output):
        """Get the function that can advance the sampler state.

        This function is called by the MCMC sampler to draw and return a new sample.

        Args:
            nmr_samples (int): the number of samples we will draw
            thinning (int): the thinning factor we want to use
            return_output (boolean): if the kernel should return output

        Returns:
            str: a CL function with signature:

            .. code-block:: c
                void _advanceSampler(mot_data_struct* data,
                                     ulong current_iteration,
                                     void* rng_data,
                                     local mot_float_type* current_position,
                                     local double* const current_likelihood,
                                     local mot_float_type* const current_prior,
                                     local double* log_likelihood_tmp);
        """
        raise NotImplementedError()

    def _get_log_prior_cl_func(self):
        """Get the CL log prior compute function.

        Returns:
            str: the compute function for computing the log prior of the model.
        """
        prior_func = self._model.get_log_prior_function()
        return SimpleCLFunction.from_string('''
            mot_float_type _computeLogPrior(mot_data_struct* data,
                                            local const mot_float_type* const x){
                return ''' + prior_func.get_cl_function_name() + '''(data, x);
            }
        ''', dependencies=[prior_func])

    def _get_log_likelihood_cl_func(self):
        """Get the CL log likelihood compute function.

        This uses local reduction to compute the log likelihood for every observation in CL local space.
        The results are then summed in the first work item and returned using a pointer.

        Returns:
            str: the CL code for the log likelihood compute func.
        """
        ll_func = self._model.get_log_likelihood_function()
        return SimpleCLFunction.from_string('''
            double _computeLogLikelihood(mot_data_struct* data,
                                       local mot_float_type* const current_position,
                                       local double* log_likelihood_tmp){
                
                return ''' + ll_func.get_cl_function_name() + '''(data, current_position, log_likelihood_tmp);
            }
        ''', dependencies=[ll_func])

    @contextmanager
    def _logging(self, nmr_samples, burnin, thinning):
        self._logger.info('Starting sample with method {0}'.format(self.__class__.__name__))
        self._logger.info('We will use a {} precision float type for the calculations.'.format(
            'double' if self._cl_runtime_info.double_precision else 'single'))

        for env in self._cl_runtime_info.get_cl_environments():
            self._logger.info('Using device \'{}\'.'.format(str(env)))

        self._logger.info('Using compile flags: {}'.format(self._cl_runtime_info.get_compile_flags()))

        sample_settings = dict(nmr_samples=nmr_samples,
                               burnin=burnin,
                               thinning=thinning)
        self._logger.info('Sample settings: nmr_samples: {nmr_samples}, burnin: {burnin}, '
                          'thinning: {thinning}. '.format(**sample_settings))

        samples_drawn = dict(samples_drawn=(burnin + thinning * nmr_samples),
                             samples_returned=nmr_samples)
        self._logger.info('Total samples drawn: {samples_drawn}, total samples returned: '
                          '{samples_returned} (per problem).'.format(**samples_drawn))
        yield
        self._logger.info('Finished sample')


class AbstractRWMSampler(AbstractSampler):

    def __init__(self, model, starting_positions, proposal_stds, use_random_scan=False, **kwargs):
        """An abstract basis for Random Walk Metropolis (RWM) samplers.

        Random Walk Metropolis (RWM) samplers require for every parameter and every modeling instance an proposal
        standard deviation, used in the random walk.

        Args:
            model (SampleModelInterface): the model to sample.
            starting_positions (ndarray): the starting positions for the sampler. Should be a two dimensional matrix
                with for every modeling instance (first dimension) and every parameter (second dimension) a value.
            proposal_stds (ndarray): for every parameter and every modeling instance an initial proposal std.
            use_random_scan (boolean): if we iterate over the parameters in a random order or in a linear order
                at every sample iteration. By default we apply a system scan starting from the first dimension to the
                last. With a random scan we randomize the indices every iteration.
        """
        super(AbstractRWMSampler, self).__init__(model, starting_positions, **kwargs)
        self._proposal_stds = np.copy(np.require(proposal_stds, requirements='CAOW',
                                                 dtype=self._cl_runtime_info.mot_float_dtype))
        self._use_random_scan = use_random_scan

    def _get_kernel_data(self, nmr_samples, thinning, return_output):
        kernel_data = super(AbstractRWMSampler, self)._get_kernel_data(nmr_samples, thinning, return_output)
        kernel_data.update({
            '_proposal_stds': Array(self._proposal_stds, 'mot_float_type', is_writable=True,
                                    ensure_zero_copy=True)
        })
        return kernel_data

    def _get_proposal_update_function(self, nmr_samples, thinning, return_output):
        """Get the proposal update function.

        Returns:
            str: the update function for the proposal standard deviations.
                Should be a string with CL code with the signature:

                .. code-block:: c
                    void _updateProposalState(mot_data_struct* data,
                                              ulong current_iteration,
                                              local mot_float_type* current_position);
        """
        return '''
            void _updateProposalState(mot_data_struct* data, ulong current_iteration){}
        '''

    def _at_acceptance_callback_c_func(self):
        """Get a CL function that is to be applied at the moment a sample is accepted.

        Returns:
            str: a piece of C code to be applied when the function is accepted.
                Should be a string with CL code with the signature:

                .. code-block:: c
                    void _sampleAccepted(mot_data_struct* data,
                                         ulong current_iteration,
                                         uint parameter_ind);
        """
        return '''
            void _sampleAccepted(mot_data_struct* data, ulong current_iteration, uint parameter_ind){}
        '''

    def _get_state_update_cl_func(self, nmr_samples, thinning, return_output):
        proposal_finalize_func = self._model.get_finalize_proposal_function()

        kernel_source = self._get_proposal_update_function(nmr_samples, thinning, return_output)
        kernel_source += self._at_acceptance_callback_c_func()
        kernel_source += proposal_finalize_func.get_cl_code()

        if self._use_random_scan:
            kernel_source += '''
                void _shuffle(uint* array, uint n, void* rng_data){
                    if(n > 1){
                        for(uint i = 0; i < n - 1; i++){
                          uint j = (uint)(frand(rng_data) * (n - i) + i); 
                          
                          uint tmp = array[j];
                          array[j] = array[i];
                          array[i] = tmp;
                        }
                    }
                }
            '''

        kernel_source += '''
            void _advanceSampler(
                    mot_data_struct* data,
                    ulong current_iteration, 
                    void* rng_data,
                    local mot_float_type* current_position,
                    local double* const current_likelihood,
                    local mot_float_type* const current_prior,
                    local double* log_likelihood_tmp){

                local mot_float_type new_position[''' + str(self._nmr_params) + '''];
                local mot_float_type new_prior;
                local double new_likelihood;
                local double bayesian_f;
                bool is_first_work_item = get_local_id(0) == 0;

                if(is_first_work_item){
                    for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                        new_position[k] = current_position[k];
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);     
        '''
        if self._use_random_scan:
            kernel_source += '''
                uint indices[] = {''' + ', '.join(map(str, range(self._nmr_params))) + '''};
                _shuffle(indices, ''' + str(self._nmr_params) + ''', rng_data);
                
                for(uint ind = 0; ind < ''' + str(self._nmr_params) + '''; ind++){
                    uint k = indices[ind];    
            '''
        else:
            kernel_source += 'for(uint k = 0; k < ' + str(self._nmr_params) + '; k++){'
        kernel_source += '''
                    if(is_first_work_item){
                        new_position[k] += frandn(rng_data) * data->_proposal_stds[k];
                        ''' + proposal_finalize_func.get_cl_function_name() + '''(data, new_position);
                        new_prior = _computeLogPrior(data, new_position);
                    }
                    barrier(CLK_LOCAL_MEM_FENCE);

                    if(exp(new_prior) > 0){
                        new_likelihood = _computeLogLikelihood(data, new_position, log_likelihood_tmp);

                        if(is_first_work_item){
                            bayesian_f = exp((new_likelihood + new_prior) - (*current_likelihood + *current_prior));

                            if(frand(rng_data) < bayesian_f){
                                *current_likelihood = new_likelihood;
                                *current_prior = new_prior;
                                for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                                    current_position[k] = new_position[k];
                                }           
                                _sampleAccepted(data, current_iteration, k);
                            }
                            else{
                                for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                                    new_position[k] = current_position[k];
                                }
                            }
                        }
                    }
                    else{ // prior returned 0
                        if(is_first_work_item){
                            for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                                new_position[k] = current_position[k];
                            }
                        }
                    }
                }
                
                if(is_first_work_item){
                    _updateProposalState(data, current_iteration, current_position);
                }
                barrier(CLK_LOCAL_MEM_FENCE);
            }
        '''
        return kernel_source


class SamplingOutput(object):

    def get_samples(self):
        """Get the matrix containing the sample results.

        Returns:
            ndarray: the sampled parameter maps, a (d, p, n) array with for d problems and p parameters n samples.
        """
        raise NotImplementedError()

    def get_log_likelihoods(self):
        """Get per set of sampled parameters the log likelihood value associated with that set of parameters.

        Returns:
            ndarray: the log likelihood values, a (d, n) array with for d problems and n samples the log likelihood
                value.
        """
        raise NotImplementedError()

    def get_log_priors(self):
        """Get per set of sampled parameters the log prior value associated with that set of parameters.

        Returns:
            ndarray: the log prior values, a (d, n) array with for d problems and n samples the prior value.
        """
        raise NotImplementedError()


class SimpleSampleOutput(SamplingOutput):

    def __init__(self, samples, log_likelihoods, log_priors):
        """Simple storage container for the sample output"""
        self._samples = samples
        self._log_likelihood = log_likelihoods
        self._log_prior = log_priors

    def get_samples(self):
        return self._samples

    def get_log_likelihoods(self):
        return self._log_likelihood

    def get_log_priors(self):
        return self._log_prior



from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Array, Zeros, LocalMemory

__author__ = 'Robbert Harms'
__date__ = "2014-05-21"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def compute_log_likelihood(model, parameters, cl_runtime_info=None):
    """Calculate and return the log likelihood of the given model for the given parameters.

    This calculates the log likelihoods for every problem in the model (typically after optimization),
    or a log likelihood for every sample of every model (typically after sample). In the case of the first (after
    optimization), the parameters must be an (d, p) array for d problems and p parameters. In the case of the
    second (after sample), you must provide this function with a matrix of shape (d, p, n) with d problems,
    p parameters and n samples.

    Args:
        model (AbstractModel): The model to calculate the full log likelihood for.
        parameters (ndarray): The parameters to use in the evaluation of the model. This is either an (d, p) matrix
            or (d, p, n) matrix with d problems, p parameters and n samples.
        cl_runtime_info (mot.lib.cl_runtime_info.CLRuntimeInfo): the runtime information

    Returns:
        ndarray: per problem the log likelihood, or, per problem and per sample the log likelihood.
    """

    def get_cl_function():
        ll_func = model.get_log_likelihood_function()
        nmr_params = parameters.shape[1]

        if len(parameters.shape) > 2:
            return SimpleCLFunction.from_string('''
                void compute(mot_data_struct* data){
                    local mot_float_type x[''' + str(nmr_params) + '''];

                    for(uint sample_ind = 0; sample_ind < ''' + str(parameters.shape[2]) + '''; sample_ind++){
                        for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                            x[i] = data->parameters[i *''' + str(parameters.shape[2]) + ''' + sample_ind];
                        }
                        
                        double ll = ''' + ll_func.get_cl_function_name() + '''(data, x, data->local_reduction_lls);
                        if(get_local_id(0) == 0){
                            *(data->log_likelihoods) = ll;
                        }
                    }
                }
            ''', dependencies=[ll_func])

        return SimpleCLFunction.from_string('''
            void compute(mot_data_struct* data){
                local mot_float_type x[''' + str(nmr_params) + '''];
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    x[i] = data->parameters[i];
                }
                
                double ll = ''' + ll_func.get_cl_function_name() + '''(data, x, data->local_reduction_lls);
                if(get_local_id(0) == 0){
                    *(data->log_likelihoods) = ll;
                }
            }
        ''', dependencies=[ll_func])

    all_kernel_data = dict(model.get_kernel_data())
    all_kernel_data['parameters'] = Array(parameters)

    shape = parameters.shape
    if len(shape) > 2:
        all_kernel_data.update({
            'log_likelihoods': Zeros((shape[0], shape[2]), 'mot_float_type'),
            'local_reduction_lls': LocalMemory('double')
        })
    else:
        all_kernel_data.update({
            'log_likelihoods': Zeros((shape[0],), 'mot_float_type'),
            'local_reduction_lls': LocalMemory('double')
        })

    get_cl_function().evaluate({'data': all_kernel_data}, nmr_instances=parameters.shape[0], use_local_reduction=True,
                               cl_runtime_info=cl_runtime_info)

    return all_kernel_data['log_likelihoods'].get_data()


def compute_objective_value(model, parameters, cl_runtime_info=None):
    """Calculate and return the objective function value of the given model for the given parameters.

    Args:
        model (AbstractModel): The model to calculate the objective function for
        parameters (ndarray): The parameters to use in the evaluation of the model, an (d, p) matrix
            with d problems and p parameters.
        cl_runtime_info (mot.lib.cl_runtime_info.CLRuntimeInfo): the runtime information

    Returns:
        ndarray: vector matrix with per problem the objective function value
    """
    objective_func = model.get_objective_function()
    nmr_params = parameters.shape[1]

    cl_function = SimpleCLFunction.from_string('''
        void compute(mot_data_struct* data){
            local mot_float_type x[''' + str(nmr_params) + '''];

            for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                x[i] = data->parameters[i];
            }

            double objective = ''' + objective_func.get_cl_function_name() + '''(
                data, x, 0, data->local_reduction_lls);

            if(get_local_id(0) == 0){
                *(data->objective_values) = objective;
            }
       }
   ''', dependencies=[objective_func])

    all_kernel_data = dict(model.get_kernel_data())
    all_kernel_data.update({
        'parameters': Array(parameters),
        'objective_values': Zeros((parameters.shape[0],), 'mot_float_type'),
        'local_reduction_lls': LocalMemory('double')
    })

    cl_function.evaluate({'data': all_kernel_data}, nmr_instances=parameters.shape[0],
                         use_local_reduction=True, cl_runtime_info=cl_runtime_info)

    return all_kernel_data['objective_values'].get_data()




from mot.sample import AdaptiveMetropolisWithinGibbs
from mot.sample import SingleComponentAdaptiveMetropolis

__author__ = 'Robbert Harms'
__date__ = "2015-07-06"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


samplers = [AdaptiveMetropolisWithinGibbs, SingleComponentAdaptiveMetropolis]


def get_sampler_by_name(name):
    """ Get the class by the given name.

    This does not instantiate the class, only returns a reference to it.

    Args:
        name: the name of the optimizer we want to return

    Returns:
        class: the class of the sampler requested
    """
    return _get_item(name, samplers, 'samplers')


def _get_item(name, item_list, factory_type):
    for item in item_list:
        if item.__name__ == name:
            return item
    raise ValueError('The item with the name {0} could not be found in the {1} factory.'.format(name, factory_type))



from .amwg import AdaptiveMetropolisWithinGibbs
from .scam import SingleComponentAdaptiveMetropolis

__author__ = 'Robbert Harms'
__date__ = "2014-05-21"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"



import tatsu

__author__ = 'Robbert Harms'
__date__ = "2015-03-21"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


_cl_data_type_parser = tatsu.compile('''
    result = [address_space] {pre_type_qualifiers}* data_type [{post_type_qualifiers}*];

    address_space = ['__'] ('local' | 'global' | 'constant' | 'private');
    pre_type_qualifiers = 'const' | 'volatile';
    
    data_type = type_specifier [vector_length] {pointer}*;
    pointer = '*';
    vector_length = /[2348(16]/;
    
    post_type_qualifiers = 'const' | 'restrict';
    
    # match some as regex, to ensure the vector length is taken into account.
    type_specifier =
                     ?'mot_float_type'
                   | 'mot_data_struct'
                   | 'void'
                   | ?'bool'
                   | ?'char' | ?'cl_char' | ?'unsigned char' | ?'uchar' | ?'cl_uchar'
                   | ?'short' | ?'cl_short' | ?'unsigned short' | ?'ushort'
                   | ?'int' | ?'unsigned int' | ?'uint'
                   | ?'long' | ?'unsigned long' | ?'ulong'
                   | ?'half'
                   | ?'float'
                   | ?'double';
''')


class CLDataType(object):
    """Interface for CL data type containers.

    Basically this encapsulates the type and its qualifiers that define a variable in CL.
    """

    def get_declaration(self):
        """Get the complete CL declaration for this datatype.

        Returns:
            str: the declaration for this data type.
        """
        raise NotImplementedError()

    @property
    def declaration_type(self):
        """Get the type of this parameter in CL language

        This only returns the parameter type (like ``double`` or ``int*`` or ``float4*`` ...). It does not include other
        qualifiers.

        Returns:
            str: The name of this data type
        """
        raise NotImplementedError()

    @property
    def raw_data_type(self):
        """Get the raw data type without the vector and pointer additions.

        For example, if the data type is float4*, we will only return float here.

        Returns:
            str: the raw CL data type
        """
        raise NotImplementedError()

    @property
    def ctype(self):
        """Get the ctype of this data type.

        For example, if the data type is float4*, we will return float4 here.

        Returns:
            str: the full ctype of this data type
        """
        raise NotImplementedError()

    @property
    def is_vector_type(self):
        """Check if this data type is a vector type (like for example double4, float2, int8, etc.).

        Returns:
            boolean: True if it is a vector type, false otherwise
        """
        raise NotImplementedError()

    @property
    def is_pointer_type(self):
        """Check if this parameter is a pointer type (appended by a ``*``)

        Returns:
            boolean: True if it is a pointer type, false otherwise
        """
        raise NotImplementedError()

    @property
    def nmr_pointers(self):
        """Get the number of asterisks / pointer references of this data type.

        If the data type is float**, we return 2 here.

        Returns:
            int: the number of pointer asterisks in the data type.
        """
        raise NotImplementedError()

    @property
    def vector_length(self):
        """Get the length of this vector, returns None if not a vector type.

        Returns:
            int: the length of the vector type (for example, if the data type is float4, this returns 4).
        """
        raise NotImplementedError()

    @property
    def address_space(self):
        """Get the address space of this data declaration.

        Returns:
            str: the data type address space, one of ``global``, ``local``, ``constant`` or ``private``.
        """
        raise NotImplementedError()


class SimpleCLDataType(CLDataType):

    def __init__(self, raw_data_type, nmr_pointers=0, vector_length=None, address_space=None,
                 pre_type_qualifiers=None, post_type_qualifiers=None):
        """Create a new CL data type container.

        The CL type can either be a CL native type (``half``, ``double``, ``int``, ...) or the
        special ``mot_float_type`` type.

        Args:
            raw_data_type (str): the specific data type without the vector number and asterisks
            nmr_pointers (int): The number of dereferences of this parameter (number of ``*``).
            vector_length (int or None): If None this data type is not a CL vector type.
                If it is an integer it is the vector length of this data type (2, 3, 4, ...)
            address_space (str or None): the address space qualifier or None if not used. One of:
                {``__local``, ``local``, ``__global``, ``global``,
                ``__constant``, ``constant``, ``__private``, ``private``} or None.
            pre_type_qualifiers (Union[List[str], None]): the type qualifiers to use before the type.
                One or more of {const, volatile}
            post_type_qualifiers (Union[List[str], None]): the type qualifiers to use before the type.
                One or more of {const, restrict}
        """
        self._raw_data_type = str(raw_data_type)
        self._nmr_pointers = nmr_pointers
        self._vector_length = vector_length

        if self.vector_length:
            self._vector_length = int(self.vector_length)

        self._address_space = address_space

        if self._address_space is not None:
            if '__' in self._address_space:
                self._address_space = self._address_space[2:]

            valid_address_spaces = ('global', 'constant', 'local', 'private')

            if self._address_space not in valid_address_spaces:
                raise ValueError('The given address space qualifier "{}" is not one of {}.'.format(
                    self._address_space, valid_address_spaces))

        self.pre_type_qualifiers = pre_type_qualifiers
        if isinstance(self.pre_type_qualifiers, str):
            self.pre_type_qualifiers = [self.pre_type_qualifiers]

        self.post_type_qualifiers = post_type_qualifiers
        if isinstance(self.post_type_qualifiers, str):
            self.post_type_qualifiers = [post_type_qualifiers]

    @classmethod
    def from_string(cls, parameter_declaration):
        """Parse the parameter declaration into a CLDataType

        Args:
            parameter_declaration (str): the CL parameter declaration. Example: ``global const float4*`` const

        Returns:
            SimpleCLDataType: the CL data type for this parameter declaration
        """
        class Semantics(object):

            def __init__(self):
                self._raw_data_type = None
                self._nmr_pointers = 0
                self._vector_length = None
                self._address_space = None
                self._pre_type_qualifiers = None
                self._post_type_qualifiers = None

            def result(self, ast):
                return SimpleCLDataType(
                    self._raw_data_type,
                    nmr_pointers=self._nmr_pointers,
                    vector_length=self._vector_length,
                    address_space=self._address_space,
                    pre_type_qualifiers=self._pre_type_qualifiers,
                    post_type_qualifiers=self._post_type_qualifiers)

            def pointer(self, ast):
                self._nmr_pointers += 1
                return ast

            def vector_length(self, ast):
                self._vector_length = int(ast)
                return ast

            def type_specifier(self, ast):
                self._raw_data_type = ast
                return ast

            def address_space(self, ast):
                self._address_space = ''.join(ast)
                return ast

            def pre_type_qualifiers(self, ast):
                self._pre_type_qualifiers = ast
                return ast

            def post_type_qualifiers(self, ast):
                self._post_type_qualifiers = ast
                return ast

        return _cl_data_type_parser.parse(parameter_declaration, semantics=Semantics())

    def get_declaration(self):
        declaration = ''
        if self._address_space:
            declaration += str(self._address_space) + ' '
        if self.pre_type_qualifiers:
            declaration += str(' '.join(self.pre_type_qualifiers)) + ' '
        declaration += str(self.declaration_type)
        if self.post_type_qualifiers:
            declaration += ' ' + str(' '.join(self.post_type_qualifiers)) + ' '
        return declaration

    @property
    def declaration_type(self):
        s = self.raw_data_type

        if self.vector_length is not None:
            s += str(self.vector_length)

        if self.is_pointer_type:
            s += '*' * self._nmr_pointers

        return str(s)

    @property
    def is_vector_type(self):
        return self.vector_length is not None

    @property
    def is_pointer_type(self):
        return self._nmr_pointers > 0

    @property
    def nmr_pointers(self):
        return self._nmr_pointers

    @property
    def vector_length(self):
        return self._vector_length

    @property
    def raw_data_type(self):
        return self._raw_data_type

    @property
    def ctype(self):
        if self.is_vector_type:
            return self._raw_data_type + str(self.vector_length)
        return self._raw_data_type

    @property
    def address_space(self):
        if self._address_space is None:
            return 'private'
        return self._address_space

    def __str__(self):
        return self.get_declaration()



"""This module contains some diagnostic functions to diagnose the performance of MCMC sample.

The two most important functions are :func:`multivariate_ess` and :func:`univariate_ess` to calculate the effective
sample size of your samples.
"""
from collections import Mapping
import itertools
import numpy as np
from numpy.linalg import det
from scipy.special import gammaln
from scipy.stats import chi2

from mot.lib.utils import multiprocess_mapping

__author__ = 'Robbert Harms'
__date__ = "2017-03-07"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def multivariate_ess(samples, batch_size_generator=None):
    r"""Estimate the multivariate Effective Sample Size for the samples of every problem.

    This essentially applies :func:`estimate_multivariate_ess` to every problem.

    Args:
        samples (ndarray, dict or generator): either a matrix of shape (d, p, n) with d problems, p parameters and
            n samples, or a dictionary with for every parameter a matrix with shape (d, n) or, finally,
            a generator function that yields sample arrays of shape (p, n).
        batch_size_generator (MultiVariateESSBatchSizeGenerator): the batch size generator, tells us how many
            batches and of which size we use in estimating the minimum ESS.

    Returns:
        ndarray: the multivariate ESS per problem
    """
    samples_generator = _get_sample_generator(samples)
    return np.array(multiprocess_mapping(_MultivariateESSMultiProcessing(batch_size_generator), samples_generator()))


class _MultivariateESSMultiProcessing(object):

    def __init__(self, batch_size_generator):
        """Used in the function :func:`multivariate_ess` to estimate the multivariate ESS using multiprocessing."""
        self._batch_size_generator = batch_size_generator

    def __call__(self, samples):
        return estimate_multivariate_ess(samples, batch_size_generator=self._batch_size_generator)


def univariate_ess(samples, method='standard_error', **kwargs):
    r"""Estimate the univariate Effective Sample Size for the samples of every problem.

    This computes the ESS using:

        .. math::

            ESS(X) = n * \frac{\lambda^{2}}{\sigma^{2}}

        Where :math:`\lambda` is the standard deviation of the chain and :math:`\sigma` is estimated using the
        monte carlo standard error (which in turn is, by default, estimated using a batch means estimator).

    Args:
        samples (ndarray, dict or generator): either a matrix of shape (d, p, n) with d problems, p parameters and
            n samples, or a dictionary with for every parameter a matrix with shape (d, n) or, finally,
            a generator function that yields sample arrays of shape (p, n).
        method (str): one of 'autocorrelation' or 'standard_error' defaults to 'standard_error'.
            If 'autocorrelation' is chosen we apply the function: :func:`estimate_univariate_ess_autocorrelation`,
            if 'standard_error` is choosen we apply the function: :func:`estimate_univariate_ess_standard_error`.
        **kwargs: passed to the chosen compute method

    Returns:
        ndarray: a matrix of size (d, p) with for every problem and every parameter an ESS.

    References:
        * Flegal, J.M., Haran, M., and Jones, G.L. (2008). "Markov chain Monte Carlo: Can We
          Trust the Third Significant Figure?". Statistical Science, 23, p. 250-260.
        * Marc S. Meketon and Bruce Schmeiser. 1984. Overlapping batch means: something for nothing?.
          In Proceedings of the 16th conference on Winter simulation (WSC '84), Sallie Sheppard (Ed.).
          IEEE Press, Piscataway, NJ, USA, 226-230.
    """
    samples_generator = _get_sample_generator(samples)
    return np.array(multiprocess_mapping(_UnivariateESSMultiProcessing(method, **kwargs), samples_generator()))


class _UnivariateESSMultiProcessing(object):

    def __init__(self, method, **kwargs):
        """Used in the function :func:`univariate_ess` to estimate the univariate ESS using multiprocessing."""
        self._method = method
        self._kwargs = kwargs

    def __call__(self, samples):
        if self._method == 'autocorrelation':
            compute_func = estimate_univariate_ess_autocorrelation
        else:
            compute_func = estimate_univariate_ess_standard_error

        result = np.zeros(samples.shape[0])
        for param_ind in range(samples.shape[0]):
            result[param_ind] = compute_func(samples[param_ind], **self._kwargs)

        return result


def _get_sample_generator(samples):
    """Get a sample generator from the given polymorphic input.

    Args:
        samples (ndarray, dict or generator): either an matrix of shape (d, p, n) with d problems, p parameters and
            n samples, or a dictionary with for every parameter a matrix with shape (d, n) or, finally,
            a generator function that yields sample arrays of shape (p, n).

    Returns:
        generator: a generator that yields a matrix of size (p, n) for every problem in the input.
    """
    if isinstance(samples, Mapping):
        def samples_generator():
            for ind in range(samples[list(samples.keys())[0]].shape[0]):
                yield np.array([samples[s][ind, :] for s in sorted(samples)])
    elif isinstance(samples, np.ndarray):
        def samples_generator():
            for ind in range(samples.shape[0]):
                yield samples[ind]
    else:
        samples_generator = samples
    return samples_generator


def get_auto_correlation(chain, lag):
    r"""Estimates the auto correlation for the given chain (1d vector) with the given lag.

    Given a lag :math:`k`, the auto correlation coefficient :math:`\rho_{k}` is estimated as:

    .. math::

        \hat{\rho}_{k} = \frac{E[(X_{t} - \mu)(X_{t + k} - \mu)]}{\sigma^{2}}

    Please note that this equation only works for lags :math:`k < n` where :math:`n` is the number of samples in
    the chain.

    Args:
        chain (ndarray): the vector with the samples
        lag (int): the lag to use in the autocorrelation computation

    Returns:
        float: the autocorrelation with the given lag
    """
    normalized_chain = chain - np.mean(chain, dtype=np.float64)
    lagged_mean = np.mean(normalized_chain[:len(chain) - lag] * normalized_chain[lag:], dtype=np.float64)
    return lagged_mean / np.var(chain, dtype=np.float64)


def get_auto_correlation_time(chain, max_lag=None):
    r"""Compute the auto correlation time up to the given lag for the given chain (1d vector).

    This will halt when the maximum lag :math:`m` is reached or when the sum of two consecutive lags for any
    odd lag is lower or equal to zero.

    The auto correlation sum is estimated as:

    .. math::

        \tau = 1 + 2 * \sum_{k=1}^{m}{\rho_{k}}


    Where :math:`\rho_{k}` is estimated as:

    .. math::

        \hat{\rho}_{k} = \frac{E[(X_{t} - \mu)(X_{t + k} - \mu)]}{\sigma^{2}}

    Args:
        chain (ndarray): the vector with the samples
        max_lag (int): the maximum lag to use in the autocorrelation computation. If not given we use:
            :math:`min(n/3, 1000)`.
    """
    max_lag = max_lag or min(len(chain) // 3, 1000)

    normalized_chain = chain - np.mean(chain, dtype=np.float64)

    previous_accoeff = 0
    auto_corr_sum = 0

    for lag in range(1, max_lag):
        auto_correlation_coeff = np.mean(normalized_chain[:len(chain) - lag] * normalized_chain[lag:], dtype=np.float64)

        if lag % 2 == 0:
            if previous_accoeff + auto_correlation_coeff <= 0:
                break

        auto_corr_sum += auto_correlation_coeff
        previous_accoeff = auto_correlation_coeff

    return auto_corr_sum / np.var(chain, dtype=np.float64)


def estimate_univariate_ess_autocorrelation(chain, max_lag=None):
    r"""Estimate effective sample size (ESS) using the autocorrelation of the chain.

    The ESS is an estimate of the size of an iid sample with the same variance as the current sample.
    This function implements the ESS as described in Kass et al. (1998) and Robert and Casella (2004; p. 500):

    .. math::

        ESS(X) = \frac{n}{\tau} = \frac{n}{1 + 2 * \sum_{k=1}^{m}{\rho_{k}}}

    where :math:`\rho_{k}` is estimated as:

    .. math::

        \hat{\rho}_{k} = \frac{E[(X_{t} - \mu)(X_{t + k} - \mu)]}{\sigma^{2}}

    References:
        * Kass, R. E., Carlin, B. P., Gelman, A., and Neal, R. (1998)
            Markov chain Monte Carlo in practice: A roundtable discussion. The American Statistician, 52, 93--100.
        * Robert, C. P. and Casella, G. (2004) Monte Carlo Statistical Methods. New York: Springer.
        * Geyer, C. J. (1992) Practical Markov chain Monte Carlo. Statistical Science, 7, 473--483.

    Args:
        chain (ndarray): the chain for which to calculate the ESS, assumes a vector of length ``n`` samples
        max_lag (int): the maximum lag used in the variance calculations. If not given defaults to
            :math:`min(n/3, 1000)`.

    Returns:
        float: the estimated ESS
    """
    return len(chain) / (1 + 2 * get_auto_correlation_time(chain, max_lag))


def estimate_univariate_ess_standard_error(chain, batch_size_generator=None, compute_method=None):
    r"""Compute the univariate ESS using the standard error method.

    This computes the ESS using:

    .. math::

        ESS(X) = n * \frac{\lambda^{2}}{\sigma^{2}}

    Where :math:`\lambda` is the standard deviation of the chain and :math:`\sigma` is estimated using the monte carlo
    standard error (which in turn is, by default, estimated using a batch means estimator).

    Args:
        chain (ndarray): the Markov chain
        batch_size_generator (UniVariateESSBatchSizeGenerator): the method that generates that batch sizes
            we will use. Per default it uses the :class:`SquareRootSingleBatch` method.
        compute_method (ComputeMonteCarloStandardError): the method used to compute the standard error.
            By default we will use the :class:`BatchMeansMCSE` method

    Returns:
        float: the estimated ESS
    """
    sigma = (monte_carlo_standard_error(chain, batch_size_generator=batch_size_generator,
                                        compute_method=compute_method) ** 2 * len(chain))
    lambda_ = np.var(chain, dtype=np.float64)
    return len(chain) * (lambda_ / sigma)


def minimum_multivariate_ess(nmr_params, alpha=0.05, epsilon=0.05):
    r"""Calculate the minimum multivariate Effective Sample Size you will need to obtain the desired precision.

    This implements the inequality from Vats et al. (2016):

    .. math::

        \widehat{ESS} \geq \frac{2^{2/p}\pi}{(p\Gamma(p/2))^{2/p}} \frac{\chi^{2}_{1-\alpha,p}}{\epsilon^{2}}

    Where :math:`p` is the number of free parameters.

    Args:
        nmr_params (int): the number of free parameters in the model
        alpha (float): the level of confidence of the confidence region. For example, an alpha of 0.05 means
            that we want to be in a 95% confidence region.
        epsilon (float): the level of precision in our multivariate ESS estimate.
            An epsilon of 0.05 means that we expect that the Monte Carlo error is 5% of the uncertainty in
            the target distribution.

    Returns:
        float: the minimum multivariate Effective Sample Size that one should aim for in MCMC sample to
            obtain the desired confidence region with the desired precision.

    References:
        Vats D, Flegal J, Jones G (2016). Multivariate Output Analysis for Markov Chain Monte Carlo.
        arXiv:1512.07713v2 [math.ST]
    """
    tmp = 2.0 / nmr_params
    log_min_ess = tmp * np.log(2) + np.log(np.pi) - tmp * (np.log(nmr_params) + gammaln(nmr_params / 2)) \
                  + np.log(chi2.ppf(1 - alpha, nmr_params)) - 2 * np.log(epsilon)
    return int(round(np.exp(log_min_ess)))


def multivariate_ess_precision(nmr_params, multi_variate_ess, alpha=0.05):
    r"""Calculate the precision given your multivariate Effective Sample Size.

    Given that you obtained :math:`ESS` multivariate effective samples in your estimate you can calculate the
    precision with which you approximated your desired confidence region.

    This implements the inequality from Vats et al. (2016), slightly restructured to give :math:`\epsilon` back instead
    of the minimum ESS.

    .. math::

         \epsilon = \sqrt{\frac{2^{2/p}\pi}{(p\Gamma(p/2))^{2/p}} \frac{\chi^{2}_{1-\alpha,p}}{\widehat{ESS}}}

    Where :math:`p` is the number of free parameters and ESS is the multivariate ESS from your samples.

    Args:
        nmr_params (int): the number of free parameters in the model
        multi_variate_ess (int): the number of iid samples you obtained in your sample results.
        alpha (float): the level of confidence of the confidence region. For example, an alpha of 0.05 means
            that we want to be in a 95% confidence region.

    Returns:
        float: the minimum multivariate Effective Sample Size that one should aim for in MCMC sample to
            obtain the desired confidence region with the desired precision.

    References:
        Vats D, Flegal J, Jones G (2016). Multivariate Output Analysis for Markov Chain Monte Carlo.
        arXiv:1512.07713v2 [math.ST]
    """
    tmp = 2.0 / nmr_params
    log_min_ess = tmp * np.log(2) + np.log(np.pi) - tmp * (np.log(nmr_params) + gammaln(nmr_params / 2)) \
                  + np.log(chi2.ppf(1 - alpha, nmr_params)) - np.log(multi_variate_ess)
    return np.sqrt(np.exp(log_min_ess))


def estimate_multivariate_ess_sigma(samples, batch_size):
    r"""Calculates the Sigma matrix which is part of the multivariate ESS calculation.

    This implementation is based on the Matlab implementation found at: https://github.com/lacerbi/multiESS

    The Sigma matrix is defined as:

    .. math::

        \Sigma = \Lambda + 2 * \sum_{k=1}^{\infty}{Cov(Y_{1}, Y_{1+k})}

    Where :math:`Y` are our samples and :math:`\Lambda` is the covariance matrix of the samples.

    This implementation computes the :math:`\Sigma` matrix using a Batch Mean estimator using the given batch size.
    The batch size has to be :math:`1 \le b_n \le n` and a typical value is either :math:`\lfloor n^{1/2} \rfloor`
    for slow mixing chains or :math:`\lfloor n^{1/3} \rfloor` for reasonable mixing chains.

    If the length of the chain is longer than the sum of the length of all the batches, this implementation
    calculates :math:`\Sigma` for every offset and returns the average of those offsets.

    Args:
        samples (ndarray): the samples for which we compute the sigma matrix. Expects an (p, n) array with
            p the number of parameters and n the sample size
        batch_size (int): the batch size used in the approximation of the correlation covariance

    Returns:
        ndarray: an pxp array with p the number of parameters in the samples.

    References:
        Vats D, Flegal J, Jones G (2016). Multivariate Output Analysis for Markov Chain Monte Carlo.
        arXiv:1512.07713v2 [math.ST]
    """
    sample_means = np.mean(samples, axis=1, dtype=np.float64)
    nmr_params, chain_length = samples.shape

    nmr_batches = int(np.floor(chain_length / batch_size))
    sigma = np.zeros((nmr_params, nmr_params))

    nmr_offsets = chain_length - nmr_batches * batch_size + 1

    for offset in range(nmr_offsets):
        batches = np.reshape(samples[:, np.array(offset + np.arange(0, nmr_batches * batch_size), dtype=np.int)].T,
                             [batch_size, nmr_batches, nmr_params], order='F')

        batch_means = np.squeeze(np.mean(batches, axis=0, dtype=np.float64))

        Z = batch_means - sample_means

        for x, y in itertools.product(range(nmr_params), range(nmr_params)):
            sigma[x, y] += np.sum(Z[:, x] * Z[:, y])

    return sigma * batch_size / (nmr_batches - 1) / nmr_offsets


def estimate_multivariate_ess(samples, batch_size_generator=None, full_output=False):
    r"""Compute the multivariate Effective Sample Size of your (single instance set of) samples.

    This multivariate ESS is defined in Vats et al. (2016) and is given by:

    .. math::

        ESS = n \bigg(\frac{|\Lambda|}{|\Sigma|}\bigg)^{1/p}

    Where :math:`n` is the number of samples, :math:`p` the number of parameters, :math:`\Lambda` is the covariance
    matrix of the parameters and :math:`\Sigma` captures the covariance structure in the target together with
    the covariance due to correlated samples. :math:`\Sigma` is estimated using
    :func:`estimate_multivariate_ess_sigma`.

    In the case of NaN in any part of the computation the ESS is set to 0.

    To compute the multivariate ESS for multiple problems, please use :func:`multivariate_ess`.

    Args:
        samples (ndarray): an pxn matrix with for p parameters and n samples.
        batch_size_generator (MultiVariateESSBatchSizeGenerator): the batch size generator, tells us how many
            batches and of which size we use for estimating the minimum ESS. Defaults to :class:`SquareRootSingleBatch`
        full_output (boolean): set to True to return the estimated :math:`\Sigma` and the optimal batch size.

    Returns:
        float or tuple: when full_output is set to True we return a tuple with the estimated multivariate ESS,
            the estimated :math:`\Sigma` matrix and the optimal batch size. When full_output is False (the default)
            we only return the ESS.

    References:
        Vats D, Flegal J, Jones G (2016). Multivariate Output Analysis for Markov Chain Monte Carlo.
        arXiv:1512.07713v2 [math.ST]
    """
    batch_size_generator = batch_size_generator or SquareRootSingleBatch()

    batch_sizes = batch_size_generator.get_multivariate_ess_batch_sizes(*samples.shape)

    nmr_params, chain_length = samples.shape
    nmr_batches = len(batch_sizes)

    det_lambda = det(np.cov(samples))

    ess_estimates = np.zeros(nmr_batches)
    sigma_estimates = np.zeros((nmr_params, nmr_params, nmr_batches))

    for i in range(0, nmr_batches):
        sigma = estimate_multivariate_ess_sigma(samples, int(batch_sizes[i]))
        ess = chain_length * (det_lambda**(1.0 / nmr_params) / det(sigma)**(1.0 / nmr_params))

        ess_estimates[i] = ess
        sigma_estimates[..., i] = sigma

    ess_estimates = np.nan_to_num(ess_estimates)

    if nmr_batches > 1:
        idx = np.argmin(ess_estimates)
    else:
        idx = 0

    if full_output:
        return ess_estimates[idx], sigma_estimates[..., idx], batch_sizes[idx]
    return ess_estimates[idx]


def monte_carlo_standard_error(chain, batch_size_generator=None, compute_method=None):
    """Compute Monte Carlo standard errors for the expectations

    This is a convenience function that calls the compute method for each batch size and returns the lowest ESS
    over the used batch sizes.

    Args:
        chain (ndarray): the Markov chain
        batch_size_generator (UniVariateESSBatchSizeGenerator): the method that generates that batch sizes
            we will use. Per default it uses the :class:`SquareRootSingleBatch` method.
        compute_method (ComputeMonteCarloStandardError): the method used to compute the standard error.
            By default we will use the :class:`BatchMeansMCSE` method
    """
    batch_size_generator = batch_size_generator or SquareRootSingleBatch()
    compute_method = compute_method or BatchMeansMCSE()

    batch_sizes = batch_size_generator.get_univariate_ess_batch_sizes(len(chain))

    return np.min(list(compute_method.compute_standard_error(chain, b) for b in batch_sizes))


class MultiVariateESSBatchSizeGenerator(object):
    """Objects of this class are used as input to the multivariate ESS function.

    The multivariate ESS function needs to have at least one batch size to use during the computations. More batch
    sizes are also possible and the batch size with the lowest ESS is then preferred.
    Objects of this class implement the logic behind choosing batch sizes.
    """

    def get_multivariate_ess_batch_sizes(self, nmr_params, chain_length):
        r"""Get the batch sizes to use for the calculation of the Effective Sample Size (ESS).

        This should return a list of batch sizes that the ESS calculation will use to determine :math:`\Sigma`

        Args:
            nmr_params (int): the number of parameters in the samples
            chain_length (int): the length of the chain

        Returns:
            list: the batches of the given sizes we will test in the ESS calculations
        """


class UniVariateESSBatchSizeGenerator(object):
    """Objects of this class are used as input to the univariate ESS function that uses the batch means.

    The univariate batch means ESS function needs to have at least one batch size to use during the computations.
    More batch sizes are also possible and the batch size with the lowest ESS is then preferred.
    Objects of this class implement the logic behind choosing batch sizes.
    """

    def get_univariate_ess_batch_sizes(self, chain_length):
        r"""Get the batch sizes to use for the calculation of the univariate Effective Sample Size (ESS).

        This should return a list of batch sizes that the ESS calculation will use to determine :math:`\sigma`

        Args:
            chain_length (int): the length of the chain

        Returns:
            list: the batches of the given sizes we will test in the ESS calculations
        """


class SquareRootSingleBatch(MultiVariateESSBatchSizeGenerator, UniVariateESSBatchSizeGenerator):
    r"""Returns :math:`\sqrt(n)`."""

    def get_multivariate_ess_batch_sizes(self, nmr_params, chain_length):
        return [np.floor(chain_length**(1/2.0))]

    def get_univariate_ess_batch_sizes(self, chain_length):
        return [np.floor(chain_length ** (1 / 2.0))]


class CubeRootSingleBatch(MultiVariateESSBatchSizeGenerator, UniVariateESSBatchSizeGenerator):
    r"""Returns :math:`n^{1/3}`."""

    def get_multivariate_ess_batch_sizes(self, nmr_params, chain_length):
        return [np.floor(chain_length**(1/3.0))]

    def get_univariate_ess_batch_sizes(self, chain_length):
        return [np.floor(chain_length ** (1 / 3.0))]


class LinearSpacedBatchSizes(MultiVariateESSBatchSizeGenerator):

    def __init__(self, nmr_batches=200):
        r"""Returns a number of batch sizes from which the ESS algorithm will select the one with the lowest ESS.

        This is a conservative choice since the lowest ESS of all batch sizes is chosen.

        The batch sizes are generated as linearly spaced values in:

        .. math::

            \Big[ n^{1/4}, max(\lfloor x/max(20,p) \rfloor, \lfloor \sqrt{n} \rfloor) \Big]

        where :math:`n` is the chain length and :math:`p` is the number of parameters.

        Args:
            nmr_batches (int): the number of linearly spaced batches we will generate.
        """
        self._nmr_batches = nmr_batches

    def get_multivariate_ess_batch_sizes(self, nmr_params, chain_length):
        b_min = np.floor(chain_length**(1 / 4.0))
        b_max = np.max((np.floor(chain_length / np.max((nmr_params, 20))),
                        np.floor(chain_length**(1 / 2.0))))
        return list(np.unique(np.round(np.exp(np.linspace(np.log(b_min), np.log(b_max), self._nmr_batches)))))


class ComputeMonteCarloStandardError(object):
    """Method to compute the Monte Carlo Standard error."""

    def compute_standard_error(self, chain, batch_size):
        """Compute the standard error of the given chain and the given batch size.

        Args:
            chain (ndarray): the chain for which to compute the SE
            batch_size (int): batch size or window size to use in the computations

        Returns:
            float: the Monte Carlo Standard Error
        """
        raise NotImplementedError()


class BatchMeansMCSE(ComputeMonteCarloStandardError):
    """Computes the Monte Carlo Standard Error using simple batch means."""

    def compute_standard_error(self, chain, batch_size):
        nmr_batches = int(np.floor(len(chain) / batch_size))

        chain_mean = np.mean(chain, dtype=np.float64)
        var_sum = 0

        for batch_index in range(nmr_batches):
            batch_mean = np.mean(chain[int(batch_index * batch_size):int((batch_index + 1) * batch_size)],
                                 dtype=np.float64)
            var_sum += (batch_mean - chain_mean)**2

        var_hat = batch_size * var_sum / (nmr_batches - 1)
        return np.sqrt(var_hat / len(chain))


class OverlappingBatchMeansMCSE(ComputeMonteCarloStandardError):
    """Computes the Monte Carlo Standard Error using overlapping batch means."""

    def compute_standard_error(self, chain, batch_size):
        nmr_batches = int(len(chain) - batch_size + 1)

        chain_mean = np.mean(chain, dtype=np.float64)
        var_sum = 0

        for batch_index in range(nmr_batches):
            batch_mean = np.mean(chain[int(batch_index):int(batch_index + batch_size)], dtype=np.float64)
            var_sum += (batch_mean - chain_mean) ** 2

        var_hat = (len(chain) * batch_size * var_sum) / (nmr_batches - 1) / nmr_batches
        return np.sqrt(var_hat / len(chain))



import itertools
import numpy as np

from mot.lib.cl_function import SimpleCLFunction
from mot.lib.cl_runtime_info import CLRuntimeInfo
from mot.lib.kernel_data import LocalMemory, Array, Zeros
from scipy import linalg


__author__ = 'Robbert Harms'
__date__ = '2017-10-16'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class NumericalHessian(object):

    def __init__(self, cl_runtime_info=None):
        """Base class for CL routines.

        Args:
            cl_runtime_info (mot.lib.cl_runtime_info.CLRuntimeInfo): the runtime information
        """
        self._cl_runtime_info = cl_runtime_info or CLRuntimeInfo()

    def set_cl_runtime_info(self, cl_runtime_info):
        """Update the CL runtime information.

        Args:
            cl_runtime_info (mot.lib.cl_runtime_info.CLRuntimeInfo): the new runtime information
        """
        self._cl_runtime_info = cl_runtime_info

    def calculate(self, model, parameters, step_ratio=2, nmr_steps=15, step_offset=None):
        """Calculate and return the Hessian of the given function at the given parameters.

        This calculates the Hessian using central difference (using a 2nd order Taylor expansion) with a Richardson
        extrapolation over the proposed sequence of steps.

        The Hessian is evaluated at the steps:

        .. math::
            \quad  ((f(x + d_j e_j + d_k e_k) - f(x + d_j e_j - d_k e_k)) -
                    (f(x - d_j e_j + d_k e_k) - f(x - d_j e_j - d_k e_k)) /
                    (4 d_j d_k)

        where :math:`e_j` is a vector where element :math:`j` is one and the rest are zero
        and :math:`d_j` is a scalar spacing :math:`steps_j`.

        Steps are generated according to a exponentially diminishing ratio defined as:

            steps = max_step * step_ratio**-(i+offset), i=0, 1,.., nmr_steps-1.

        Where the max step is taken from the model. For example, a maximum step of 2 with a step ratio of 2 and with
        4 steps gives: [2.0, 1.0, 0.5, 0.25]. If offset would be 2, we would instead get: [0.5, 0.25, 0.125, 0.0625].

        If number of steps is 1, we use not Richardson extrapolation and return the results of the first step. If the
        number of steps is 2 we use a first order Richardson extrapolation step. For all higher number of steps
        we use a second order Richardson extrapolation.

        Args:
            model (mot.lib.model_interfaces.NumericalDerivativeInterface): the model to use for calculating the
                derivative.
            parameters (ndarray): The parameters at which to evaluate the gradient. A (d, p) matrix with d problems,
                p parameters and n samples.
            step_ratio (float): the ratio at which the steps diminish.
            nmr_steps (int): the number of steps we will generate. We will calculate the derivative for each of these
                step sizes and extrapolate the best step size from among them. The minimum number of steps is 2.
            step_offset (the offset in the steps, if set we start the steps from the given offset.

        Returns:
            ndarray: the gradients for each of the parameters for each of the problems
        """
        if len(parameters.shape) == 1:
            parameters = parameters[None, :]
        nmr_params = parameters.shape[1]

        parameter_scalings = np.array(model.numdiff_get_scaling_factors())

        def finalize_derivatives(derivatives):
            """Transforms the derivatives from vector to matrix and apply the parameter scalings."""
            return self._results_vector_to_matrix(derivatives, nmr_params) \
                   * np.outer(parameter_scalings, parameter_scalings)

        derivatives = self._compute_derivatives(model, parameters, step_ratio, step_offset, nmr_steps)

        if nmr_steps == 1:
            return finalize_derivatives(derivatives[..., 0])

        derivatives, errors = self._richardson_extrapolation(derivatives, step_ratio)

        if nmr_steps <= 3:
            return finalize_derivatives(derivatives[..., 0])

        if derivatives.shape[2] > 2:
            derivatives, errors = self._wynn_extrapolate(derivatives)

        if derivatives.shape[2] == 1:
            return finalize_derivatives(derivatives[..., 0])

        derivatives, errors = self._median_outlier_extrapolation(derivatives, errors)
        return finalize_derivatives(derivatives)

    def _compute_derivatives(self, model, parameters, step_ratio, step_offset, nmr_steps):
        """Compute the lower triangular elements of the Hessian using the central difference method.

        This will compute the elements of the Hessian multiple times with decreasing step sizes.

        Args:
            model: the log likelihood model we are trying to differentiate
            parameters (ndarray): a (n, p) matrix with for for every problem n, p parameters. These are the points
                at which we want to calculate the derivative
            step_ratio (float): the ratio at which the steps exponentially diminish
            step_offset (int): ignore the first few step sizes by this offset
            nmr_steps (int): the number of steps to compute and return (after the step offset)
        """
        parameter_scalings = np.array(model.numdiff_get_scaling_factors())

        nmr_params = parameters.shape[1]
        nmr_derivatives = (nmr_params ** 2 - nmr_params) // 2 + nmr_params

        initial_step = self._get_initial_step_size(model, parameters)
        if step_offset:
            initial_step *= float(step_ratio) ** -step_offset

        all_kernel_data = dict(model.get_kernel_data())
        all_kernel_data.update({
            'parameters': Array(parameters, ctype='mot_float_type'),
            'local_reduction_lls': LocalMemory('double'),
            'parameter_scalings_inv': Array(1. / parameter_scalings, ctype='float', offset_str='0'),
            'initial_step': Array(initial_step, ctype='float'),
            'step_evaluates': Zeros((parameters.shape[0], nmr_derivatives, nmr_steps), 'double'),
        })

        self._derivation_kernel(model, nmr_params, nmr_steps, step_ratio).evaluate(
            {'data': all_kernel_data}, nmr_instances=parameters.shape[0], use_local_reduction=True)

        return all_kernel_data['step_evaluates'].get_data()

    def _richardson_extrapolation(self, derivatives, step_ratio):
        """Apply the Richardson extrapolation to the derivatives computed with different steps.

        Having for every problem instance and every Hessian element multiple derivatives computed with decreasing steps,
        we can now apply Richardson extrapolation to reduce the error term from :math:`\mathcal{O}(h^{2})` to
        :math:`\mathcal{O}(h^{4})` or :math:`\mathcal{O}(h^{6})` depending on how many steps we have calculated.

        This method only considers extrapolation up to the sixth error order. For a set of two derivatives we compute
        a single fourth order approximation, for three derivatives and up we compute ``n-2`` sixth order approximations.
        Expected errors for approximation ``i`` are computed using the ``i+1`` derivative plus a statistical error
        based on the machine precision.

        Args:
            derivatives (ndarray): (n, p, s), a matrix with for n problems and p parameters, s step sizes.
            step_ratio (ndarray): the diminishing ratio of the steps used to compute the derivatives.
        """
        nmr_problems, nmr_derivatives, nmr_steps = derivatives.shape
        richardson_coefficients = self._get_richardson_coefficients(step_ratio, min(nmr_steps, 3) - 1)
        nmr_convolutions_needed = nmr_steps - (len(richardson_coefficients) - 2)
        final_nmr_convolutions = nmr_convolutions_needed - 1

        kernel_data = {
            'derivatives': Array(derivatives, 'double', offset_str='{problem_id} * ' + str(nmr_steps)),
            'richardson_extrapolations': Zeros(
                (nmr_problems * nmr_derivatives, nmr_convolutions_needed), 'double', is_readable=True),
            'errors': Zeros(
                (nmr_problems * nmr_derivatives, final_nmr_convolutions), 'double', is_readable=True),
        }

        richardson_func = self._richardson_error_kernel(nmr_steps, nmr_convolutions_needed, richardson_coefficients)
        richardson_func.evaluate({'data': kernel_data}, nmr_instances=nmr_problems * nmr_derivatives,
                                 use_local_reduction=False, cl_runtime_info=self._cl_runtime_info)

        richardson_extrapolations = np.reshape(kernel_data['richardson_extrapolations'].get_data(),
                                               (nmr_problems, nmr_derivatives, nmr_convolutions_needed))
        errors = np.reshape(kernel_data['errors'].get_data(),
                            (nmr_problems, nmr_derivatives, final_nmr_convolutions))

        return richardson_extrapolations[..., :final_nmr_convolutions], errors

    def _wynn_extrapolate(self, derivatives):
        nmr_problems, nmr_derivatives, nmr_steps = derivatives.shape
        nmr_extrapolations = nmr_steps - 2

        kernel_data = {
            'derivatives': Array(derivatives, 'double', offset_str='{problem_id} * ' + str(nmr_steps)),
            'extrapolations': Zeros((nmr_problems * nmr_derivatives, nmr_extrapolations),
                                                         'double', is_readable=True),
            'errors': Zeros((nmr_problems * nmr_derivatives, nmr_extrapolations), 'double',
                            is_readable=True),
        }

        wynn_func = self._wynn_extrapolation_kernel(nmr_steps)
        wynn_func.evaluate({'data': kernel_data}, nmr_instances=nmr_problems * nmr_derivatives,
                           use_local_reduction=False, cl_runtime_info=self._cl_runtime_info)

        extrapolations = np.reshape(kernel_data['extrapolations'].get_data(),
                                    (nmr_problems, nmr_derivatives, nmr_extrapolations))
        errors = np.reshape(kernel_data['errors'].get_data(), (nmr_problems, nmr_derivatives, nmr_extrapolations))
        return extrapolations, errors

    def _median_outlier_extrapolation(self, derivatives, errors):
        """Add an error to outliers and afterwards return the derivatives with the lowest errors.

        This seems to be the slowest function in the library. Perhaps one day update this to OpenCL as well.
        Some ideas are in: http://krstn.eu/np.nanpercentile()-there-has-to-be-a-faster-way/
        """
        def _get_median_outliers_errors(der, trim_fact=10):
            """Discards any estimate that differs wildly from the median of the estimates (of that derivative).

            A factor of 10 to 1 in either direction . The actual trimming factor is
            defined as a parameter.
            """
            p25, median, p75 = np.nanpercentile(der, q=[25, 50, 75], axis=2)[..., None]
            iqr = np.abs(p75 - p25)

            a_median = np.abs(median)
            outliers = (((abs(der) < (a_median / trim_fact)) +
                         (abs(der) > (a_median * trim_fact))) * (a_median > 1e-8) +
                        ((der < p25 - 1.5 * iqr) + (p75 + 1.5 * iqr < der)))
            return outliers * np.abs(der - median)

        all_nan = np.where(np.sum(np.isnan(errors), axis=2) == errors.shape[2])
        errors[all_nan[0], all_nan[1], 0] = 0
        derivatives[all_nan[0], all_nan[1], 0] = 0

        errors += _get_median_outliers_errors(derivatives)

        minpos = np.nanargmin(errors, axis=2)
        indices = np.indices(minpos.shape)

        derivatives_final = derivatives[indices[0], indices[1], minpos]
        errors_final = errors[indices[0], indices[1], minpos]

        derivatives_final[all_nan[0], all_nan[1]] = np.nan
        errors_final[all_nan[0], all_nan[1]] = np.nan

        return derivatives_final, errors_final

    def _results_vector_to_matrix(self, vectors, nmr_params):
        """Transform for every problem (and optionally every step size) the vector results to a square matrix.

        Since we only process the lower triangular items, we have to convert the 1d vectors into 2d matrices for every
        problem. This function does that.

        Args:
            vectors (ndarray): for every problem (and step size) the 1d vector with results
            nmr_params (int): the number of parameters
        """
        matrices = np.zeros(vectors.shape[:-1] + (nmr_params, nmr_params), dtype=vectors.dtype)

        ltr_ind = 0
        for px in range(nmr_params):
            matrices[..., px, px] = vectors[..., ltr_ind]
            ltr_ind += 1

            for py in range(px + 1, nmr_params):
                matrices[..., px, py] = vectors[..., ltr_ind]
                matrices[..., py, px] = vectors[..., ltr_ind]
                ltr_ind += 1
        return matrices

    def _get_initial_step_size(self, model, parameters):
        """Get an initial step size to use for every parameter.

        This chooses the maximum of the maximum step defined in the model and the minimum step possible
        for a parameter given its bounds.

        Args:
            model (mot.lib.model_interfaces.NumericalDerivativeInterface): the model to use for calculating the
                derivative.
            parameters (ndarray): The parameters at which to evaluate the gradient. A (d, p) matrix with d problems,
                p parameters and n samples.

        Returns:
            ndarray: for every problem instance the vector with the initial step size for each parameter.
        """
        upper_bounds = model.get_upper_bounds()
        lower_bounds = model.get_lower_bounds()

        max_step = model.numdiff_get_max_step()

        initial_step = np.zeros_like(parameters)

        for ind in range(parameters.shape[1]):
            if model.numdiff_use_bounds()[ind]:
                use_lower = model.numdiff_use_lower_bounds()[ind]
                use_upper = model.numdiff_use_upper_bounds()[ind]

                if use_upper and not use_lower:
                    minimum_allowed_step = np.abs(upper_bounds[ind] - parameters[:, ind]) \
                                           * model.numdiff_get_scaling_factors()[ind]
                elif use_lower and not use_upper:
                    minimum_allowed_step = np.abs(parameters[:, ind] - lower_bounds[ind]) \
                                           * model.numdiff_get_scaling_factors()[ind]
                else:
                    minimum_allowed_step = np.minimum(np.abs(parameters[:, ind] - lower_bounds[ind]),
                                                      np.abs(upper_bounds[ind] - parameters[:, ind])) \
                                           * model.numdiff_get_scaling_factors()[ind]
                initial_step[:, ind] = np.minimum(minimum_allowed_step, max_step[ind])
            else:
                initial_step[:, ind] = max_step[ind]

        return initial_step

    def _get_richardson_coefficients(self, step_ratio, nmr_extrapolations):
        """Get the matrix with the convolution rules.

        In the kernel we will extrapolate the sequence based on the Richardsons method.

        This method assumes we have a series expansion like:

            L = f(h) + a0 * h^p_0 + a1 * h^p_1+ a2 * h^p_2 + ...

        where p_i = order + step * i  and f(h) -> L as h -> 0, but f(0) != L.

        If we evaluate the right hand side for different stepsizes h we can fit a polynomial to that sequence
        of approximations.

        Instead of using all the generated points at the same time, we convolute the Richardson method over the
        acquired steps to approximate the higher order error terms.

        Args:
            step_ratio (float): the ratio at which the steps diminish.
            nmr_extrapolations (int): the number of extrapolations we want to do. Each extrapolation requires
                an evaluation at an exponentially decreasing step size.

        Returns:
            ndarray: a vector with the extrapolation coefficients.
        """
        if nmr_extrapolations == 0:
            return np.array([1])

        error_diminishing_per_step = 2
        taylor_expansion_order = 2

        def r_matrix(num_terms):
            i, j = np.ogrid[0:num_terms + 1, 0:num_terms]
            r_mat = np.ones((num_terms + 1, num_terms + 1))
            r_mat[:, 1:] = (1.0 / step_ratio) ** (i * (error_diminishing_per_step * j + taylor_expansion_order))
            return r_mat

        return linalg.pinv(r_matrix(nmr_extrapolations))[0]

    def _get_compute_functions_cl(self, model, nmr_params, nmr_steps, step_ratio):
        objective_func = model.get_objective_function()
        numdiff_param_transform = model.numdiff_parameter_transformation()

        func = objective_func.get_cl_code()
        func += numdiff_param_transform.get_cl_code()

        return func + '''
            double _calculate_function(mot_data_struct* data, local mot_float_type* x){
                return ''' + objective_func.get_cl_function_name() + '''(data, x, 0, data->local_reduction_lls);
            }
            
            /**
             * Evaluate the model with a perturbation in two dimensions.
             *
             * Args:
             *  data: the data container
             *  x_input: the array with the input parameters
             *  perturb_dim_0: the index (into the x_input parameters) of the first parameter to perturbate
             *  perturb_0: the added perturbation of the index corresponding to ``perturb_dim_0``
             *  perturb_dim_1: the index (into the x_input parameters) of the second parameter to perturbate
             *  perturb_1: the added perturbation of the index corresponding to ``perturb_dim_1``
             *
             * Returns:
             *  the function evaluated at the parameters plus their perturbation.
             */
            double _eval_step(mot_data_struct* data, local mot_float_type* x_input,
                              uint perturb_dim_0, mot_float_type perturb_0,
                              uint perturb_dim_1, mot_float_type perturb_1){

                local mot_float_type x_tmp[''' + str(nmr_params) + '''];
                
                if(get_local_id(0) == 0){
                    for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x_tmp[i] = x_input[i];
                    }
                    x_tmp[perturb_dim_0] += perturb_0;
                    x_tmp[perturb_dim_1] += perturb_1;
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                ''' + numdiff_param_transform.get_cl_function_name() + '''(data, x_tmp);
                return _calculate_function(data, x_tmp);
            }
            
            /**
             * Compute one element of the Hessian for a number of steps.
             * 
             * This uses the initial steps in the data structure, indexed by the parameters to change (px, py).
             */
            void _compute_steps(mot_data_struct* data, local mot_float_type* x_input, mot_float_type f_x_input,
                                uint px, uint py, global double* step_evaluates){
                
                double step_x;
                double step_y;
                double tmp;
                bool is_first_workitem = get_local_id(0) == 0;
                
                if(px == py){
                    for(uint step_ind = 0; step_ind < ''' + str(nmr_steps) + '''; step_ind++){
                        step_x = data->initial_step[px] / pown(''' + str(float(step_ratio)) + ''', step_ind);
                        
                        tmp = (
                              _eval_step(data, x_input,
                                         px, 2 * (step_x * data->parameter_scalings_inv[px]),
                                         0, 0)
                            + _eval_step(data, x_input,
                                         px, -2 * (step_x * data->parameter_scalings_inv[px]),
                                         0, 0)
                            - 2 * f_x_input
                        ) / (4 * step_x * step_x);
                        
                        if(is_first_workitem){
                            step_evaluates[step_ind] = tmp;
                        }
                    }
                }
                else{
                    for(uint step_ind = 0; step_ind < ''' + str(nmr_steps) + '''; step_ind++){
                        step_x = data->initial_step[px] / pown(''' + str(float(step_ratio)) + ''', step_ind);
                        step_y = data->initial_step[py] / pown(''' + str(float(step_ratio)) + ''', step_ind);
                        
                        tmp = (
                              _eval_step(data, x_input,
                                         px, step_x * data->parameter_scalings_inv[px],
                                         py, step_y * data->parameter_scalings_inv[py])
                            - _eval_step(data, x_input,
                                         px, step_x * data->parameter_scalings_inv[px],
                                         py, -step_y * data->parameter_scalings_inv[py])
                            - _eval_step(data, x_input,
                                         px, -step_x * data->parameter_scalings_inv[px],
                                         py, step_y * data->parameter_scalings_inv[py])
                            + _eval_step(data, x_input,
                                         px, -step_x * data->parameter_scalings_inv[px],
                                         py, -step_y * data->parameter_scalings_inv[py])
                        ) / (4 * step_x * step_y);
    
                        if(is_first_workitem){
                            step_evaluates[step_ind] = tmp;
                        }                       
                    }
                }
            }
        '''

    def _get_error_estimate_functions_cl(self, nmr_steps, nmr_convolutions,
                                         richardson_coefficients):
        func = '''
            /**
             * Apply a simple kernel convolution over the results from each row of steps.
             *
             * This applies a convolution starting from the starting step index that contained a valid move.
             * It uses the mode 'reflect' to deal with outside points.
             *
             * Please note that this kernel is hard coded to work with 2nd order Taylor expansions derivatives only.

             * Args:
             *  step_evaluates: the step evaluates for each row of the step sizes
             *  richardson_extrapolations: the array to place the convoluted results in
             */
            void _apply_richardson_convolution(
                    global double* derivatives,
                    global double* richardson_extrapolations){

                double convolution_kernel[''' + str(len(richardson_coefficients)) + '''] = {''' + \
                    ', '.join(map(str, richardson_coefficients)) + '''};
                
                for(uint step_ind = 0; step_ind < ''' + str(nmr_convolutions) + '''; step_ind++){
                    
                    // convolute the Richardson coefficients
                    for(uint kernel_ind = 0; kernel_ind < ''' + str(len(richardson_coefficients)) + '''; kernel_ind++){
                        
                        uint kernel_step_ind = step_ind + kernel_ind;

                        // reflect
                        if(kernel_step_ind >= ''' + str(nmr_steps) + '''){
                            kernel_step_ind -= 2 * (kernel_step_ind - ''' + str(nmr_steps) + ''') + 1;
                        }

                        richardson_extrapolations[step_ind] +=
                            derivatives[kernel_step_ind] * convolution_kernel[kernel_ind];
                    }
                }
            }
            
            /**
             * Compute the errors from using the Richardson extrapolation.
             *
             * "A neat trick to compute the statistical uncertainty in the estimate of our desired derivative is to use 
             *  statistical methodology for that error estimate. While I do appreciate that there is nothing truly 
             *  statistical or stochastic in this estimate, the approach still works nicely, providing a very 
             *  reasonable estimate in practice. A three term Richardson-like extrapolant, then evaluated at four 
             *  distinct values for \delta, will yield an estimate of the standard error of the constant term, with one 
             *  spare degree of freedom. The uncertainty is then derived by multiplying that standard error by the 
             *  appropriate percentile from the Students-t distribution." 
             * 
             * Cited from https://numdifftools.readthedocs.io/en/latest/src/numerical/derivest.html
             * 
             * In addition to the error derived from the various estimates, we also approximate the numerical round-off 
             * errors in this method. All in all, the resulting errors should reflect the absolute error of the
             * estimates.
             */ 
            void _compute_richardson_errors(
                    global double* derivatives, 
                    global double* richardson_extrapolations,
                    global double* errors){
                
                // The magic number 12.7062... follows from the student T distribution with one dof. 
                //  >>> import scipy.stats as ss
                //  >>> allclose(ss.t.cdf(12.7062047361747, 1), 0.975) # True
                double fact = max(
                    (mot_float_type)''' + str(12.7062047361747 * np.sqrt(np.sum(richardson_coefficients**2))) + ''',
                    (mot_float_type)MOT_EPSILON * 10);
                
                double tolerance;
                double error;
                
                for(uint conv_ind = 0; conv_ind < ''' + str(nmr_convolutions - 1) + '''; conv_ind++){
                    tolerance = max(fabs(richardson_extrapolations[conv_ind + 1]), 
                                    fabs(richardson_extrapolations[conv_ind])
                                    ) * MOT_EPSILON * fact;
                    
                    error = fabs(richardson_extrapolations[conv_ind] - richardson_extrapolations[conv_ind + 1]) * fact;
                    
                    if(error <= tolerance){
                        error += tolerance * 10;
                    }
                    else{
                        error += fabs(richardson_extrapolations[conv_ind] - 
                                      derivatives[''' + str(nmr_steps - nmr_convolutions + 1) + ''' + conv_ind]) * fact;
                    }
                    
                    errors[conv_ind] = error;
                }
            }
        '''
        return func

    def _derivation_kernel(self, model, nmr_params, nmr_steps, step_ratio):
        coords = [(x, y) for x, y in itertools.combinations_with_replacement(range(nmr_params), 2)]
        func = self._get_compute_functions_cl(model, nmr_params, nmr_steps, step_ratio)

        return SimpleCLFunction.from_string('''
            void compute(mot_data_struct* data){
                local mot_float_type x_input[''' + str(nmr_params) + '''];
                
                if(get_local_id(0) == 0){
                    for(uint param_ind = 0; param_ind < ''' + str(nmr_params) + '''; param_ind++){
                        x_input[param_ind] = data->parameters[param_ind];
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
                
                double f_x_input = _calculate_function(data, x_input);
                
                uint coords[''' + str(len(coords)) + '''][2] = {
                    ''' + ', '.join('{{{}, {}}}'.format(*c) for c in coords)  + '''
                };
                
                for(uint coord_ind = 0; coord_ind < ''' + str(len(coords)) + '''; coord_ind++){
                    _compute_steps(data, x_input, f_x_input, coords[coord_ind][0], coords[coord_ind][1], 
                                   data->step_evaluates + coord_ind * ''' + str(nmr_steps) + ''');
                }
            }
        ''', cl_extra=func)

    def _richardson_error_kernel(self, nmr_steps, nmr_convolutions, richardson_coefficients):
        func = self._get_error_estimate_functions_cl(nmr_steps, nmr_convolutions, richardson_coefficients)
        return SimpleCLFunction.from_string('''
            void convolute(mot_data_struct* data){
                _apply_richardson_convolution(data->derivatives, data->richardson_extrapolations);
                _compute_richardson_errors(data->derivatives, data->richardson_extrapolations, data->errors);
            }
        ''', cl_extra=func)

    def _wynn_extrapolation_kernel(self, nmr_steps):
        """OpenCL kernel for extrapolating a slowly convergent sequence.

        This algorithm, known in the Python Numdifftools as DEA3, attempts to extrapolate nonlinearly to a better
        estimate of the sequence's limiting value, thus improving the rate of convergence. The routine is based on the
        epsilon algorithm of P. Wynn [1].

        References:
        - [1] C. Brezinski and M. Redivo Zaglia (1991)
                "Extrapolation Methods. Theory and Practice", North-Holland.

        - [2] C. Brezinski (1977)
                "Acceleration de la convergence en analyse numerique",
                "Lecture Notes in Math.", vol. 584,
                Springer-Verlag, New York, 1977.

        - [3] E. J. Weniger (1989)
                "Nonlinear sequence transformations for the acceleration of
                convergence and the summation of divergent series"
                Computer Physics Reports Vol. 10, 189 - 371
                http://arxiv.org/abs/math/0306302v1
        """
        return SimpleCLFunction.from_string('''
            void compute(mot_data_struct* data){
                double v0, v1, v2; 
                
                double delta0, delta1; 
                double err0, err1; 
                double tol0, tol1; 
                
                double ss;
                bool converged;
                
                double result;
                
                for(uint i = 0; i < ''' + str(nmr_steps - 2) + '''; i++){
                    v0 = data->derivatives[i];
                    v1 = data->derivatives[i + 1];
                    v2 = data->derivatives[i + 2];
                    
                    delta0 = v1 - v0;
                    delta1 = v2 - v1;
                    
                    err0 = fabs(delta0);
                    err1 = fabs(delta1);
                    
                    tol0 = max(fabs(v0), fabs(v1)) * MOT_EPSILON;
                    tol1 = max(fabs(v1), fabs(v2)) * MOT_EPSILON;
                    
                    // avoid division by zero and overflow
                    if(err0 < MOT_MIN){
                        delta0 = MOT_MIN;
                    }
                    if(err1 < MOT_MIN){
                        delta1 = MOT_MIN;
                    }
                    
                    ss = 1.0 / delta1 - 1.0 / delta0 + MOT_MIN;
                    converged = ((err0 <= tol0) && (err1 <= tol1)) || (fabs(ss * v1) <= 1e-3);
                    
                    result = (converged ? v2 : v1 + 1.0/ss);
                    
                    data->extrapolations[i] = result;
                    data->errors[i] = err0 + err1 + (converged ? tol1 * 10 : fabs(result - v2));
                }
            }
        ''')



from pkg_resources import resource_filename
from mot.library_functions import SimpleCLLibraryFromFile, SimpleCLLibrary

__author__ = 'Robbert Harms'
__date__ = '2018-05-12'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class CerfImWOfX(SimpleCLLibraryFromFile):
    def __init__(self):
        """Calculate the cerf."""
        super(CerfImWOfX, self).__init__(
            'void', 'cerf', [],
            resource_filename('mot', 'data/opencl/cerf/im_w_of_x.cl'))


class dawson(SimpleCLLibrary):
    def __init__(self):
        super().__init__('double dawson(double x){ return (sqrt(M_PI)/2.0) * im_w_of_x(x); }',
                         dependencies=[CerfImWOfX()])


class erfi(SimpleCLLibrary):
    def __init__(self):
        """Calculate the imaginary error function for a real argument (special case).

        Compute erfi(x) = -i erf(ix), the imaginary error function.
        """
        super().__init__(
            'double erfi(double x){ return x*x > 720 ? (x > 0 ? INFINITY : -INFINITY) : exp(x*x) * im_w_of_x(x); }',
            dependencies=[CerfImWOfX()])



VERSION = '0.5.5'

_items = VERSION.split('-')
VERSION_NUMBER_PARTS = tuple(int(i) for i in _items[0].split('.'))
if len(_items) > 1:
    VERSION_STATUS = _items[1]
else:
    VERSION_STATUS = ''
__version__ = VERSION



"""Contains the runtime configuration of MOT.

This consists of two parts, functions to get the current runtime settings and configuration actions to update these
settings. To set a new configuration, create a new :py:class:`ConfigAction` and use this within a context environment
using :py:func:`config_context`. Example:

.. code-block:: python

    from mot.configuration import RuntimeConfigurationAction, config_context

    with config_context(RuntimeConfigurationAction(...)):
        ...

"""
from contextlib import contextmanager
from copy import copy

from .lib.cl_environments import CLEnvironmentFactory
from .lib.load_balance_strategies import PreferGPU

__author__ = 'Robbert Harms'
__date__ = "2015-07-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


"""The runtime configuration, this can be overwritten at run time.

For any of the AbstractCLRoutines it holds that if no suitable defaults are given we use the ones provided by this
module. This entire module acts as a singleton containing the current runtime configuration.
"""
_config = {
    'cl_environments': CLEnvironmentFactory.smart_device_selection(),
    'load_balancer': PreferGPU(),
    'compile_flags': ['-cl-single-precision-constant', '-cl-denorms-are-zero', '-cl-mad-enable', '-cl-no-signed-zeros'],
    'compile_flags_to_disable_in_double_precision': ['-cl-single-precision-constant'],
    'ignore_kernel_compile_warnings': True,
    'double_precision': False
}


def should_ignore_kernel_compile_warnings():
    """Check if we should ignore kernel compile warnings or not.

    Returns:
        boolean: True if we should ignore the kernel compile warnings, false if not.
    """
    return _config['ignore_kernel_compile_warnings']


def get_cl_environments():
    """Get the current CL environment to use during CL calculations.

    Returns:
        list of CLEnvironment: the current list of CL environments.
    """
    return _config['cl_environments']


def set_cl_environments(cl_environments):
    """Set the current CL environments to the given list

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        cl_environments (list of CLEnvironment): the new list of CL environments.

    Raises:
        ValueError: if the list of environments is empty
    """
    if not cl_environments:
        raise ValueError('The list of CL Environments is empty.')
    _config['cl_environments'] = cl_environments


def get_load_balancer():
    """Get the current load balancer to use during CL calculations.

    Returns:
        SimpleLoadBalanceStrategy: the current load balancer to use
    """
    return _config['load_balancer']


def set_load_balancer(load_balancer):
    """Set the current CL environments to the given list

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        load_balancer (SimpleLoadBalanceStrategy): the current load balancer to use
    """
    _config['load_balancer'] = load_balancer


def get_compile_flags():
    """Get the default compile flags to use in a CL routine.

    Returns:
        list: the default list of compile flags we wish to use
    """
    return list(_config['compile_flags'])


def set_compile_flags(compile_flags):
    """Set the current compile flags.

    Args:
        compile_flags (list): the new list of compile flags
    """
    _config['compile_flags'] = compile_flags


def get_compile_flags_to_disable_in_double_precision():
    """Get the list of compile flags we want to disable when running in double precision.

    Returns:
        boolean: the list of flags we want to disable when running in double mode
    """
    return copy(_config['compile_flags_to_disable_in_double_precision'])


def set_default_proposal_update(proposal_update):
    """Set the default proposal update function to use in sample.

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        mot.model_building.parameter_functions.proposal_updates.ProposalUpdate: the new proposal update function
            to use by default if no specific one is provided.
    """
    _config['default_proposal_update'] = proposal_update


def use_double_precision():
    """Check if we run the computations in default precision or not.

    Returns:
        boolean: if we run the computations in double precision or not
    """
    return _config['double_precision']


def set_use_double_precision(double_precision):
    """Set the default use of double precision.

    Returns:
        boolean: if we use double precision by default or not
    """
    _config['double_precision'] = double_precision


@contextmanager
def config_context(config_action):
    """Creates a context in which the config action is applied and unapplies the configuration after execution.

    Args:
        config_action (ConfigAction): the configuration action to use
    """
    config_action.apply()
    yield
    config_action.unapply()


class ConfigAction(object):

    def __init__(self):
        """Defines a configuration action for use in a configuration context.

        This should define an apply and unapply function that sets and unsets the configuration options.

        The applying action needs to remember the state before the application of the action.
        """

    def apply(self):
        """Apply the current action to the current runtime configuration."""

    def unapply(self):
        """Reset the current configuration to the previous state."""


class SimpleConfigAction(ConfigAction):

    def __init__(self):
        """Defines a default implementation of a configuration action.

        This simple config implements a default ``apply()`` method that saves the current state and a default
        ``unapply()`` that restores the previous state.

        For developers, it is easiest to implement ``_apply()`` such that you do not manually need to store the old
        configuraration.
        """
        super(SimpleConfigAction, self).__init__()
        self._old_config = {}

    def apply(self):
        """Apply the current action to the current runtime configuration."""
        self._old_config = {k: v for k, v in _config.items()}
        self._apply()

    def unapply(self):
        """Reset the current configuration to the previous state."""
        for key, value in self._old_config.items():
            _config[key] = value

    def _apply(self):
        """Implement this function add apply() logic after this class saves the current config."""


class RuntimeConfigurationAction(SimpleConfigAction):

    def __init__(self, cl_environments=None, load_balancer=None, compile_flags=None, double_precision=None):
        """Updates the runtime settings.

        Args:
            cl_environments (list of CLEnvironment): the new CL environments we wish to use for future computations
            load_balancer (SimpleLoadBalanceStrategy): the load balancer to use
            compile_flags (list): the list of compile flags to use during analysis.
            double_precision (boolean): if we compute in double precision or not
        """
        super(RuntimeConfigurationAction, self).__init__()
        self._cl_environments = cl_environments
        self._load_balancer = load_balancer
        self._compile_flags = compile_flags
        self._double_precision = double_precision

    def _apply(self):
        if self._cl_environments is not None:
            set_cl_environments(self._cl_environments)

        if self._load_balancer is not None:
            set_load_balancer(self._load_balancer)

        if self._compile_flags is not None:
            set_compile_flags(self._compile_flags)

        if self._double_precision is not None:
            set_use_double_precision(self._double_precision)


class VoidConfigurationAction(ConfigAction):

    def __init__(self):
        """Does nothing, useful as a default config action.
        """
        super(VoidConfigurationAction, self).__init__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_mot
----------------------------------

Tests for `mot` module.
"""

import unittest
import numpy as np

from mot import minimize
from mot.lib.cl_function import SimpleCLFunction

from mot.lib.model_interfaces import OptimizeModelInterface


class CLRoutineTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(CLRoutineTestCase, self).__init__(*args, **kwargs)


class TestRosenbrock(CLRoutineTestCase):

    def setUp(self):
        super(TestRosenbrock, self).setUp()
        self.model = Rosenbrock(5)
        self.methods = {'Nelder-Mead': None, 'Powell': {'patience': 3}}

    def test_model(self):
        for method, options in self.methods.items():
            output = minimize(self.model, np.array([[3] * 5]), method=method, options=options)
            v = output['x']
            for ind in range(2):
                self.assertAlmostEqual(v[0, ind], 1, places=3, msg=method)


class TestLSQNonLinExample(CLRoutineTestCase):

    def setUp(self):
        super(TestLSQNonLinExample, self).setUp()
        self.model = MatlabLSQNonlinExample()
        self.methods = ('Levenberg-Marquardt', 'Powell', 'Nelder-Mead')

    def test_model(self):
        for method in self.methods:
            output = minimize(self.model, np.array([[0.3, 0.4]]), method=method)
            v = output['x']
            for ind in range(2):
                self.assertAlmostEqual(v[0, ind], 0.2578, places=3, msg=method)


class Rosenbrock(OptimizeModelInterface):

    def __init__(self, n=5):
        """When optimized the parameters should all be equal to 1."""
        super(OptimizeModelInterface, self).__init__()
        self.n = n

    @property
    def name(self):
        return 'rosenbrock'

    def get_kernel_data(self):
        return {}

    def get_nmr_observations(self):
        return self.n - 1

    def get_objective_function(self):
        return SimpleCLFunction.from_string('''
            double rosenbrock_MLE_func(mot_data_struct* data, 
                                       local const mot_float_type* const x,
                                       local mot_float_type* objective_list,
                                       local double* objective_value_tmp){

                double sum = 0;
                double eval;
                for(uint i = 0; i < ''' + str(self.get_nmr_observations()) + '''; i++){
                    eval = 100 * pown(x[i + 1] - pown(x[i], 2), 2) + pown(1 - x[i], 2);
                    sum += eval;
                    
                    if(objective_list){
                        objective_list[i] = eval;
                    }
                }
                return sum;
            }
        ''')

    def get_lower_bounds(self):
        return [-np.inf] * self.n

    def get_upper_bounds(self):
        return [np.inf] * self.n


class MatlabLSQNonlinExample(OptimizeModelInterface):

    def __init__(self):
        """When optimized the parameters should be close to [0.2578, 0.2578] or something with a similar 2 norm.

        See the matlab manual page at http://nl.mathworks.com/help/optim/ug/lsqnonlin.html for more information.
        (viewed at 2015-04-02).

        """
        super(OptimizeModelInterface, self).__init__()

    @property
    def name(self):
        return 'matlab_lsqnonlin_example'

    def get_kernel_data(self):
        return {}

    def get_nmr_observations(self):
        return 10

    def get_objective_function(self):
        return SimpleCLFunction.from_string('''
            double lsqnonlin_example_objective(mot_data_struct* data, 
                                               local const mot_float_type* const x,
                                               local mot_float_type* objective_list, 
                                               local double* objective_value_tmp){
                
                double sum = 0;
                double eval;
                for(uint i = 0; i < ''' + str(self.get_nmr_observations()) + '''; i++){
                    eval = pown(2 + 2 * (i+1) - exp((i+1) * x[0]) - exp((i+1) * x[1]), 2);
                    sum += eval;
                    
                    if(objective_list){
                        objective_list[i] = eval;
                    }
                }
                return sum;
            }
        ''')

    def get_lower_bounds(self):
        return [0, 0]

    def get_upper_bounds(self):
        return [np.inf] * 2


if __name__ == '__main__':
    unittest.main()



import logging
from .__version__ import VERSION, VERSION_STATUS, __version__
from .optimize import minimize, get_minimizer_options

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
logging.getLogger(__name__).addHandler(NullHandler())


def smart_device_selection():
    """Get a list of device environments that is suitable for use in MOT.

    Returns:
        list of CLEnvironment: List with the CL device environments.
    """
    from mot.lib.cl_environments import CLEnvironmentFactory
    return CLEnvironmentFactory.smart_device_selection()



import numpy as np
import scipy
from scipy.optimize import minimize
from scipy.stats import norm
import scipy.integrate

from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Array, Zeros, Scalar
from mot.lib.utils import is_scalar, multiprocess_mapping

__author__ = 'Robbert Harms'
__date__ = '2017-11-01'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


def fit_gaussian(samples, ddof=0):
    """Calculates the mean and the standard deviation of the given samples.

    Args:
        samples (ndarray): a one or two dimensional array. If one dimensional we calculate the fit using all
            values. If two dimensional, we fit the Gaussian for every set of samples over the first dimension.
        ddof (int): the difference degrees of freedom in the std calculation. See numpy.
    """
    if len(samples.shape) == 1:
        return np.mean(samples), np.std(samples, ddof=ddof)
    return np.mean(samples, axis=1), np.std(samples, axis=1, ddof=ddof)


def fit_circular_gaussian(samples, high=np.pi, low=0):
    """Compute the circular mean for samples in a range

    Args:
        samples (ndarray): a one or two dimensional array. If one dimensional we calculate the fit using all
            values. If two dimensional, we fit the Gaussian for every set of samples over the first dimension.
        high (float): The maximum wrap point
        low (float): The minimum wrap point
    """
    cl_func = SimpleCLFunction.from_string('''
        void compute(mot_data_struct* data){
            double cos_mean = 0;
            double sin_mean = 0;
            double ang;

            for(uint i = 0; i < data->nmr_samples; i++){
                ang = (data->samples[i] - data->low)*2*M_PI / (data->high - data->low);

                cos_mean += (cos(ang) - cos_mean) / (i + 1);
                sin_mean += (sin(ang) - sin_mean) / (i + 1);
            }

            double R = hypot(cos_mean, sin_mean);
            if(R > 1){
                R = 1;
            }

            double stds = 1/2. * sqrt(-2 * log(R));

            double res = atan2(sin_mean, cos_mean);
            if(res < 0){
                 res += 2 * M_PI;
            }

            *(data->means) = res*(data->high - data->low)/2.0/M_PI + data->low;
            *(data->stds) = ((data->high - data->low)/2.0/M_PI) * sqrt(-2*log(R));
        }
    ''')

    def run_cl(samples):
        all_kernel_data = {'samples': Array(samples, 'mot_float_type'),
                           'means': Zeros(samples.shape[0], 'mot_float_type'),
                           'stds': Zeros(samples.shape[0], 'mot_float_type'),
                           'nmr_samples': Scalar(samples.shape[1]),
                           'low': Scalar(low),
                           'high': Scalar(high),
                           }

        cl_func.evaluate({'data': all_kernel_data}, nmr_instances=samples.shape[0])
        return all_kernel_data['means'].get_data(), all_kernel_data['stds'].get_data()

    if len(samples.shape) == 1:
        mean, std = run_cl(samples[None, :])
        return mean[0], std[0]
    return run_cl(samples)


def fit_truncated_gaussian(samples, lower_bounds, upper_bounds):
    """Fits a truncated gaussian distribution on the given samples.

    This will do a maximum likelihood estimation of a truncated Gaussian on the provided samples, with the
    truncation points given by the lower and upper bounds.

    Args:
        samples (ndarray): a one or two dimensional array. If one dimensional we fit the truncated Gaussian on all
            values. If two dimensional, we calculate the truncated Gaussian for every set of samples over the
            first dimension.
        lower_bounds (ndarray or float): the lower bound, either a scalar or a lower bound per problem (first index of
            samples)
        upper_bounds (ndarray or float): the upper bound, either a scalar or an upper bound per problem (first index of
            samples)

    Returns:
        mean, std: the mean and std of the fitted truncated Gaussian
    """
    if len(samples.shape) == 1:
        return _TruncatedNormalFitter()((samples, lower_bounds, upper_bounds))

    def item_generator():
        for ind in range(samples.shape[0]):
            if is_scalar(lower_bounds):
                lower_bound = lower_bounds
            else:
                lower_bound = lower_bounds[ind]

            if is_scalar(upper_bounds):
                upper_bound = upper_bounds
            else:
                upper_bound = upper_bounds[ind]

            yield (samples[ind], lower_bound, upper_bound)

    results = np.array(multiprocess_mapping(_TruncatedNormalFitter(), item_generator()))
    return results[:, 0], results[:, 1]


def gaussian_overlapping_coefficient(means_0, stds_0, means_1, stds_1, lower=None, upper=None):
    """Compute the overlapping coefficient of two Gaussian continuous_distributions.

    This computes the :math:`\int_{-\infty}^{\infty}{\min(f(x), g(x))\partial x}` where
    :math:`f \sim \mathcal{N}(\mu_0, \sigma_0^{2})` and :math:`f \sim \mathcal{N}(\mu_1, \sigma_1^{2})` are normally
    distributed variables.

    This will compute the overlap for each element in the first dimension.

    Args:
        means_0 (ndarray): the set of means of the first distribution
        stds_0 (ndarray): the set of stds of the fist distribution
        means_1 (ndarray): the set of means of the second distribution
        stds_1 (ndarray): the set of stds of the second distribution
        lower (float): the lower limit of the integration. If not set we set it to -inf.
        upper (float): the upper limit of the integration. If not set we set it to +inf.
    """
    if lower is None:
        lower = -np.inf
    if upper is None:
        upper = np.inf

    def point_iterator():
        for ind in range(means_0.shape[0]):
            yield np.squeeze(means_0[ind]), np.squeeze(stds_0[ind]), np.squeeze(means_1[ind]), np.squeeze(stds_1[ind])

    return np.array(list(multiprocess_mapping(_ComputeGaussianOverlap(lower, upper), point_iterator())))


def deviance_information_criterions(mean_posterior_lls, ll_per_sample):
    r"""Calculates the Deviance Information Criteria (DIC) using three methods.

    This returns a dictionary returning the ``DIC_2002``, the ``DIC_2004`` and the ``DIC_Ando_2011`` method.
    The first is based on Spiegelhalter et al (2002), the second based on Gelman et al. (2004) and the last on
    Ando (2011). All cases differ in how they calculate model complexity, i.e. the effective number of parameters
    in the model. In all cases the model with the smallest DIC is preferred.

    All these DIC methods measure fitness using the deviance, which is, for a likelihood :math:`p(y | \theta)`
    defined as:

    .. math::

        D(\theta) = -2\log p(y|\theta)

    From this, the posterior mean deviance,

    .. math::

        \bar{D} = \mathbb{E}_{\theta}[D(\theta)]

    is then used as a measure of how well the model fits the data.

    The complexity, or measure of effective number of parameters, can be measured in see ways, see
    Spiegelhalter et al. (2002), Gelman et al (2004) and Ando (2011). The first method calculated the parameter
    deviance as:

    .. math::
        :nowrap:

        \begin{align}
        p_{D} &= \mathbb{E}_{\theta}[D(\theta)] - D(\mathbb{E}[\theta)]) \\
              &= \bar{D} - D(\bar{\theta})
        \end{align}

    i.e. posterior mean deviance minus the deviance evaluated at the posterior mean of the parameters.

    The second method calculated :math:`p_{D}` as:

    .. math::

        p_{D} = p_{V} = \frac{1}{2}\hat{var}(D(\theta))

    i.e. half the variance of the deviance is used as an estimate of the number of free parameters in the model.

    The third method calculates the parameter deviance as:

    .. math::

        p_{D} = 2 \cdot (\bar{D} - D(\bar{\theta}))

    That is, twice the complexity of that of the first method.

    Finally, the DIC is (for all cases) defined as:

    .. math::

        DIC = \bar{D} + p_{D}

    Args:
        mean_posterior_lls (ndarray): a 1d matrix containing the log likelihood for the average posterior
            point estimate. That is, the single log likelihood of the average parameters.
        ll_per_sample (ndarray): a (d, n) array with for d problems the n log likelihoods.
            This is the log likelihood per sample.

    Returns:
        dict: a dictionary containing the ``DIC_2002``, the ``DIC_2004`` and the ``DIC_Ando_2011`` information
            criterion maps.
    """
    mean_deviance = -2 * np.mean(ll_per_sample, axis=1)
    deviance_at_mean = -2 * mean_posterior_lls

    pd_2002 = mean_deviance - deviance_at_mean
    pd_2004 = np.var(ll_per_sample, axis=1) / 2.0

    return {'DIC_2002': np.nan_to_num(mean_deviance + pd_2002),
            'DIC_2004': np.nan_to_num(mean_deviance + pd_2004),
            'DIC_Ando_2011': np.nan_to_num(mean_deviance + 2 * pd_2002)}


class _ComputeGaussianOverlap(object):

    def __init__(self, lower, upper):
        """Helper routine for :func:`gaussian_overlapping_coefficient`.

        This calculates the overlap between two Normal distribution by taking the integral from lower to upper
        over min(f, g).
        """
        self.lower = lower
        self.upper = upper

    def __call__(self, data):
        """Compute the overlap.

        This expects data to be a tuple consisting of :math:`\mu_0, \sigma_0, \mu_1, \sigma_1`.

        Returns:
            float: the overlap between the two Gaussians.
        """
        m0, std0, m1, std1 = data

        def overlap_func(x):
            fx = norm.pdf(x, m0, std0)
            gx = norm.pdf(x, m1, std1)
            return min(fx, gx)
        return scipy.integrate.quad(overlap_func, self.lower, self.upper)[0]


class _TruncatedNormalFitter(object):

    def __call__(self, item):
        """Fit the mean and std of the truncated normal to the given samples.

        Helper function of :func:`fit_truncated_gaussian`.
        """
        samples, lower_bound, upper_bound = item
        scaling_factor = 10 ** -np.round(np.log10(np.mean(samples)))
        result = minimize(_TruncatedNormalFitter.truncated_normal_log_likelihood,
                          np.array([np.mean(samples), np.std(samples)]) * scaling_factor,
                          args=(lower_bound * scaling_factor,
                                upper_bound * scaling_factor,
                                samples * scaling_factor),
                          method='L-BFGS-B',
                          jac=_TruncatedNormalFitter.truncated_normal_ll_gradient,
                          bounds=[(lower_bound * scaling_factor, upper_bound * scaling_factor), (0, None)]
        )
        return result.x / scaling_factor

    @staticmethod
    def truncated_normal_log_likelihood(params, low, high, data):
        """Calculate the log likelihood of the truncated normal distribution.

        Args:
            params: tuple with (mean, std), the parameters under which we evaluate the model
            low (float): the lower truncation bound
            high (float): the upper truncation bound
            data (ndarray): the one dimension list of data points for which we want to calculate the likelihood

        Returns:
            float: the negative log likelihood of observing the given data under the given parameters.
                This is meant to be used in minimization routines.
        """
        mu = params[0]
        sigma = params[1]
        if sigma == 0:
            return np.inf
        ll = np.sum(norm.logpdf(data, mu, sigma))
        ll -= len(data) * np.log((norm.cdf(high, mu, sigma) - norm.cdf(low, mu, sigma)))
        return -ll

    @staticmethod
    def truncated_normal_ll_gradient(params, low, high, data):
        """Return the gradient of the log likelihood of the truncated normal at the given position.

        Args:
            params: tuple with (mean, std), the parameters under which we evaluate the model
            low (float): the lower truncation bound
            high (float): the upper truncation bound
            data (ndarray): the one dimension list of data points for which we want to calculate the likelihood

        Returns:
            tuple: the gradient of the log likelihood given as a tuple with (mean, std)
        """
        if params[1] == 0:
            return np.array([np.inf, np.inf])

        return np.array([_TruncatedNormalFitter.partial_derivative_mu(params[0], params[1], low, high, data),
                         _TruncatedNormalFitter.partial_derivative_sigma(params[0], params[1], low, high, data)])

    @staticmethod
    def partial_derivative_mu(mu, sigma, low, high, data):
        """The partial derivative with respect to the mean.

        Args:
            mu (float): the mean of the truncated normal
            sigma (float): the std of the truncated normal
            low (float): the lower truncation bound
            high (float): the upper truncation bound
            data (ndarray): the one dimension list of data points for which we want to calculate the likelihood

        Returns:
            float: the partial derivative evaluated at the given point
        """
        pd_mu = np.sum(data - mu) / sigma ** 2
        pd_mu -= len(data) * ((norm.pdf(low, mu, sigma) - norm.pdf(high, mu, sigma))
                              / (norm.cdf(high, mu, sigma) - norm.cdf(low, mu, sigma)))
        return -pd_mu

    @staticmethod
    def partial_derivative_sigma(mu, sigma, low, high, data):
        """The partial derivative with respect to the standard deviation.

        Args:
            mu (float): the mean of the truncated normal
            sigma (float): the std of the truncated normal
            low (float): the lower truncation bound
            high (float): the upper truncation bound
            data (ndarray): the one dimension list of data points for which we want to calculate the likelihood

        Returns:
            float: the partial derivative evaluated at the given point
        """
        pd_sigma = np.sum(-(1 / sigma) + ((data - mu) ** 2 / (sigma ** 3)))
        pd_sigma -= len(data) * (((low - mu) * norm.pdf(low, mu, sigma) - (high - mu) * norm.pdf(high, mu, sigma))
                                 / (sigma * (norm.cdf(high, mu, sigma) - norm.cdf(low, mu, sigma))))
        return -pd_sigma



import numpy as np
from mot.sample.base import AbstractRWMSampler
from mot.lib.kernel_data import Array

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AdaptiveMetropolisWithinGibbs(AbstractRWMSampler):

    def __init__(self, model, starting_positions, proposal_stds, target_acceptance_rate=0.44,
                 batch_size=50, damping_factor=1, min_val=1e-15, max_val=1e3, **kwargs):
        r"""An implementation of the Adaptive Metropolis-Within-Gibbs (AMWG) MCMC algorithm [1].

        This scales the proposal parameter (typically the std) such that it oscillates towards the chosen acceptance
        rate. We implement the delta function (see [1]) as: :math:`\delta(n) = \sqrt{1 / (d*n)}`.
        Where n is the current batch index and d is the damping factor. As an example, with a damping factor of 500,
        delta reaches a scaling of 0.01 in 20 batches. At a batch size of 50 that would amount to 1000 samples.

        In principal, AMWG can be used as a final MCMC algorithm if and only if the adaption is effectively zero.
        That is when delta gets close enough to zero to no longer influence the proposals.

        Args:
            model (SampleModelInterface): the model to sample.
            starting_positions (ndarray): the starting positions for the sampler. Should be a two dimensional matrix
                with for every modeling instance (first dimension) and every parameter (second dimension) a value.
            proposal_stds (ndarray): for every parameter and every modeling instance an initial proposal std.
            target_acceptance_rate (float): the target acceptance rate between 0 and 1.
            batch_size (int): the size of the batches inbetween which we update the parameters
            damping_factor (int): how fast the function moves to zero
            min_val (float): the minimum value the standard deviation can take
            max_val (float): the maximum value the standard deviation can take

        References:
            [1] Roberts GO, Rosenthal JS. Examples of adaptive MCMC. J Comput Graph Stat. 2009;18(2):349-367.
                doi:10.1198/jcgs.2009.06134.
        """
        super(AdaptiveMetropolisWithinGibbs, self).__init__(model, starting_positions, proposal_stds, **kwargs)
        self._target_acceptance_rate = target_acceptance_rate
        self._batch_size = batch_size
        self._damping_factor = damping_factor
        self._min_val = min_val
        self._max_val = max_val
        self._acceptance_counter = np.zeros((self._nmr_problems, self._nmr_params), dtype=np.uint64, order='C')

    def _get_kernel_data(self, nmr_samples, thinning, return_output):
        kernel_data = super(AdaptiveMetropolisWithinGibbs, self)._get_kernel_data(nmr_samples, thinning, return_output)
        kernel_data.update({
            '_acceptance_counter': Array(self._acceptance_counter, is_writable=True, ensure_zero_copy=True)
        })
        return kernel_data

    def _at_acceptance_callback_c_func(self):
        return '''
            void _sampleAccepted(mot_data_struct* data, ulong current_iteration, uint parameter_ind){
                data->_acceptance_counter[parameter_ind]++;
            }
        '''

    def _get_proposal_update_function(self, nmr_samples, thinning, return_output):
        kernel_source = '''
            void _updateProposalState(mot_data_struct* data, ulong current_iteration,
                                      local mot_float_type* current_position){    
                if(current_iteration > 0 && current_iteration % ''' + str(self._batch_size) + ''' == 0){
                    mot_float_type delta = sqrt(1.0/
                            (''' + str(self._damping_factor) + ''' * 
                            (current_iteration / ''' + str(self._batch_size) + ''')));
                            
                    for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                        if(data->_acceptance_counter[k] / (mot_float_type)''' + str(self._batch_size) + ''' 
                                > ''' + str(self._target_acceptance_rate) + '''){
                            data->_proposal_stds[k] *= exp(delta);
                        }
                        else{
                            data->_proposal_stds[k] /= exp(delta);
                        }
        
                        data->_proposal_stds[k] = clamp(data->_proposal_stds[k], 
                                                        (mot_float_type)''' + str(self._min_val) + ''', 
                                                        (mot_float_type)''' + str(self._max_val) + ''');
        
                        data->_acceptance_counter[k] = 0;
                    }
                }             
            }
        '''
        return kernel_source



"""The model interfaces.

This encapsulates all the information we need about models to be able to optimize or sample them using the routines
in MOT. These interfaces expose data and modeling code. The data is represented as :class:`mot.lib.utils.KernelInputData`
instances and the CL code as strings.
"""
from mot.lib.cl_function import SimpleCLFunction

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ModelBasicInfoInterface(object):

    def get_kernel_data(self):
        """Return a dictionary of input data objects we need to load into the kernel.

        Returns:
            dict[str: mot.lib.utils.KernelData]: the list of data objects we need to load in the kernel
        """
        raise NotImplementedError()

    def get_nmr_observations(self):
        """Get the number of instances/data points per problem.

        The minimum is one instance per problem.

        This number represents the number of data points

        Returns:
            int: the number of instances per problem.
        """
        raise NotImplementedError()


class OptimizeModelInterface(ModelBasicInfoInterface):

    def get_objective_function(self):
        """Get the objective function that returns the objective value and optionally, the objectives per observation.

        In the case that any of the objective lists are non-null pointers, they should be filled with the
        objective function values per observation.

        Returns:
            mot.lib.cl_function.CLFunction: A CL function with signature:

                .. code-block:: c

                    double <func_name>(mot_data_struct* data,
                                       local const mot_float_type* const x,
                                       local mot_float_type* objective_list,
                                       local double* objective_value_tmp);

                The objective list needs to be filled when the pointer is not null.
                This method should use the given local memory for the summation reduction.
        """
        raise NotImplementedError()

    def get_lower_bounds(self):
        """Get for each estimable parameter the lower bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the lower bound(s) for that parameter.
                For infinity use np.inf.
        """
        raise NotImplementedError()

    def get_upper_bounds(self):
        """Get for each estimable parameter the upper bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the upper bound(s) for that parameter.
                For infinity use np.inf.
        """
        raise NotImplementedError()


class SampleModelInterface(ModelBasicInfoInterface):

    def get_log_likelihood_function(self):
        """Get the (complete) CL Log Likelihood function that evaluates the given instance under a noise model.

        Returns:
            mot.lib.cl_function.CLFunction: A function of the kind:

                .. code-block:: c

                    double <func_name>(mot_data_struct* data,
                                       local const mot_float_type* const x,
                                       local double* objective_value_tmp);
        """
        raise NotImplementedError()

    def get_log_prior_function(self):
        """Get the prior function that returns the prior information about the given parameters.

        The prior function must be in log space.

        Returns:
            mot.lib.cl_function.CLFunction: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(
                        mot_data_struct* data,
                        local const mot_float_type* const x
                    );

            Which is called by the sample routine to calculate the prior probability.
        """
        raise NotImplementedError()

    def get_finalize_proposal_function(self):
        """Get a CL function that is called for every new proposal, in order to finalize it.

        This allows the model to change a proposal before computing the prior or likelihood probabilities.

        As an example, suppose you are sample a polar coordinate :math:`\theta` defined on
        :math:`[0, 2\pi]` with a random walk Metropolis proposal distribution. This distribution might propose positions
        outside of the range of :math:`\theta`. Of course the model function could deal with that by taking the modulus
        of the input, but then you have to post-process the chain with the same transformation. Instead, this function
        allows changing the proposal before it is put into the model and before it is stored.

        Returns:
            mot.lib.cl_function.CLFunction: A function with the signature:
                .. code-block:: c

                    void <func_name>(
                        mot_data_struct* data,
                        local mot_float_type* x
                    );

            Which is called by the sample routine to finalize the proposal.
        """
        return SimpleCLFunction(
            'void', 'finalizeProposal',
            ['mot_data_struct* data', 'local const mot_float_type* const x'],
            '')


class NumericalDerivativeInterface(OptimizeModelInterface):
    """Extends an optimization model for calculating numerical derivatives of the objective function.

    For calculating derivatives (gradients / Hessians) numerically, we need a likelihood function and some additional
    information, like the step size for each parameter, a method for checking boundary conditions and possible parameter
    transformations for circular parameters. All these extra elements are represented in this interface.
    """

    def numdiff_get_max_step(self):
        """Get for each estimable parameter the maximum step size to use for calculating numerical derivatives.

        The derivative calculation method typically uses an adaptive step size to determine the step with the best
        trade-off between numerical errors and localization of the derivative. This method must return the
        initial and largest step size to use.

        Returns:
            list[float]: per parameter a single float with the step size for that parameter
        """
        raise NotImplementedError()

    def numdiff_get_scaling_factors(self):
        """Get for each estimable parameter a scaling factor that is to be used for scaling this parameter to unitary.

        Since numerical differentiation is sensitive to differences in step sizes, it is better to rescale
        the parameters to a unitary range instead of changing the step sizes for the parameters.

        This should return numbers such that when the parameter is multiplied with this value, the magnitude of the
        parameter is about one.

        Returns:
            list[float]: per estimable parameter a single float with the parameter scaling to use for that parameter.
                Use 1 as identity.
        """
        raise NotImplementedError()

    def numdiff_use_bounds(self):
        """Check for each parameter if we should be using the bounds for that parameter when taking the derivative.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_use_upper_bounds(self):
        """Check for each parameter if we should be using the upper bounds when taking the derivative.

        This is only used if use_bounds is True for a parameter.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the upper bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_use_lower_bounds(self):
        """Check for each parameter if we should be using the lower bounds when taking the derivative.

        This is only used if use_bounds is True for a parameter.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the lower bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_parameter_transformation(self):
        """A transformation that can prepare the parameter +/- the proposed step for evaluation in the model function.

        Some parameters require for example a modulus operation before the proposed step can be used in the model
        This is the place to define them. Please note that this function does not need to incorporate boundary checks
        as those are handled already by the numerical differentiation routine.

        Returns:
            mot.lib.cl_function.CLFunction: A function with the signature:
                .. code-block:: c

                    void <func_name>(mot_data_struct* data, local mot_float_type* params);

                Where the data is the kernel data struct and params is the vector with the suggested parameters and
                which can be modified in place. Note that this is called two times, one with the parameters plus
                the step and one time without.
        """
        raise NotImplementedError()
